from datetime import datetime

from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.forms import (
	BaseForm,
	BooleanField,
	CharField,
	ChoiceField,
	DateField,
	Form,
	ImageField,
	IntegerField,
	ModelChoiceField,
	ModelForm,
)
from django.forms.utils import ErrorDict
from django.utils import timezone

from NEMO.models import (
	Account,
	Alert,
	AlertCategory,
	BuddyRequest,
	Comment,
	Consumable,
	ConsumableWithdraw,
	Project,
	ReservationItemType,
	SafetyIssue,
	ScheduledOutage,
	StaffAbsence,
	Task,
	TaskCategory,
	TaskImages,
	TemporaryPhysicalAccessRequest,
	User,
	UserPreferences,
)
from NEMO.utilities import bootstrap_primary_color, format_datetime, quiet_int
from NEMO.views.customization import get_customization


class UserForm(ModelForm):
	class Meta:
		model = User
		fields = [
			"username",
			"first_name",
			"last_name",
			"email",
			"badge_number",
			"access_expiration",
			"type",
			"domain",
			"is_active",
			"training_required",
			"physical_access_levels",
			"qualifications",
			"projects",
		]


class ProjectForm(ModelForm):
	class Meta:
		model = Project
		fields = ["name", "application_identifier", "account", "active", "start_date"]


class AccountForm(ModelForm):
	class Meta:
		model = Account
		fields = ["name", "active", "type", "start_date"]


class TaskForm(ModelForm):
	problem_category = ModelChoiceField(
		queryset=TaskCategory.objects.filter(stage=TaskCategory.Stage.INITIAL_ASSESSMENT),
		required=False,
		label="Problem category",
	)
	resolution_category = ModelChoiceField(
		queryset=TaskCategory.objects.filter(stage=TaskCategory.Stage.COMPLETION),
		required=False,
		label="Resolution category",
	)
	action = ChoiceField(choices=[("create", "create"), ("update", "update"), ("resolve", "resolve")], label="Action")
	description = CharField(required=False, label="Description")

	class Meta:
		model = Task
		fields = ["tool", "urgency", "estimated_resolution_time", "force_shutdown", "safety_hazard"]

	def __init__(self, user, *args, **kwargs):
		super(TaskForm, self).__init__(*args, **kwargs)
		self.user = user
		self.fields["tool"].required = False
		self.fields["urgency"].required = False

	def clean_description(self):
		return self.cleaned_data["description"].strip()

	def clean(self):
		if any(self.errors):
			return
		cleaned_data = super().clean()
		action = cleaned_data["action"]
		if action == "create":
			if not cleaned_data["description"]:
				raise ValidationError("You must describe the problem.")
		if action == "resolve":
			if self.instance.cancelled or self.instance.resolved:
				raise ValidationError(
					"This task can't be resolved because it is marked as 'cancelled' or 'resolved' already."
				)
		return cleaned_data

	def save(self, commit=True):
		instance = super(TaskForm, self).save(commit=False)
		action = self.cleaned_data["action"]
		description = self.cleaned_data["description"]
		instance.problem_category = self.cleaned_data["problem_category"]
		now = timezone.now()
		if action == "create":
			instance.problem_description = description
			instance.urgency = (
				Task.Urgency.HIGH
				if self.cleaned_data["force_shutdown"] or self.cleaned_data["safety_hazard"]
				else Task.Urgency.NORMAL
			)
			instance.creator = self.user
		if action == "update":
			instance.last_updated = timezone.now()
			instance.last_updated_by = self.user
			instance.cancelled = False
			instance.resolved = False
			if description:
				preface = f"On {format_datetime(now)} {self.user.get_full_name()} updated this task:\n"
				if instance.progress_description is None:
					instance.progress_description = preface + description
				else:
					instance.progress_description += "\n\n" + preface + description
				instance.progress_description = instance.progress_description.strip()
		if action == "resolve":
			instance.cancelled = False
			instance.resolved = True
			instance.resolution_time = now
			instance.resolver = self.user
			if "resolution_category" in self.cleaned_data:
				instance.resolution_category = self.cleaned_data["resolution_category"]
			if "description" in self.cleaned_data:
				if instance.resolution_description:
					preface = (
						f"On {format_datetime(now)} {self.user.get_full_name()} updated the resolution information:\n"
					)
					instance.resolution_description = (
							instance.resolution_description + "\n\n" + preface + self.cleaned_data["description"]
					).strip()
				else:
					instance.resolution_description = self.cleaned_data["description"]
		return super(TaskForm, self).save(commit=True)


class TaskImagesForm(ModelForm):
	image = ImageField(label="Images", required=False)

	class Meta:
		model = TaskImages
		fields = ("image",)


class CommentForm(ModelForm):
	class Meta:
		model = Comment
		fields = ["tool", "content", "staff_only"]

	expiration = IntegerField(label="Expiration date", min_value=0)


class SafetyIssueCreationForm(ModelForm):
	report_anonymously = BooleanField(required=False, initial=False)

	class Meta:
		model = SafetyIssue
		fields = ["reporter", "concern", "location"]

	def __init__(self, user, *args, **kwargs):
		super(SafetyIssueCreationForm, self).__init__(*args, **kwargs)
		self.user = user

	def clean_update(self):
		return self.cleaned_data["concern"].strip()

	def clean_location(self):
		return self.cleaned_data["location"].strip()

	def save(self, commit=True):
		instance = super(SafetyIssueCreationForm, self).save(commit=False)
		if not self.cleaned_data["report_anonymously"]:
			self.instance.reporter = self.user
		if commit:
			instance.save()
		return super(SafetyIssueCreationForm, self).save(commit=commit)


class SafetyIssueUpdateForm(ModelForm):
	update = CharField(required=False, label="Update")

	class Meta:
		model = SafetyIssue
		fields = ["resolved", "visible"]

	def __init__(self, user, *args, **kwargs):
		super(SafetyIssueUpdateForm, self).__init__(*args, **kwargs)
		self.user = user

	def clean_update(self):
		return self.cleaned_data["update"].strip()

	def save(self, commit=True):
		instance = super(SafetyIssueUpdateForm, self).save(commit=False)
		progress_type = "resolved" if self.cleaned_data["resolved"] else "updated"
		if progress_type == "resolved":
			instance.resolution = self.cleaned_data["update"]
			instance.resolution_time = timezone.now()
			instance.resolver = self.user
		if progress_type == "updated" and self.cleaned_data["update"]:
			progress = (
					"On "
					+ format_datetime()
					+ " "
					+ self.user.get_full_name()
					+ " updated this issue:\n"
					+ self.cleaned_data["update"]
			)
			if instance.progress:
				instance.progress += "\n\n" + progress
			else:
				instance.progress = progress
		return super(SafetyIssueUpdateForm, self).save(commit=commit)


class ConsumableWithdrawForm(ModelForm):
	class Meta:
		model = ConsumableWithdraw
		fields = ["customer", "project", "consumable", "quantity"]

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.fields["consumable"].queryset = Consumable.objects.filter(visible=True)

	def clean_customer(self):
		customer = self.cleaned_data["customer"]
		if not customer.is_active:
			raise ValidationError(
				"A consumable withdraw was requested for an inactive user. Only active users may withdraw consumables."
			)
		return customer

	def clean_project(self):
		project = self.cleaned_data["project"]
		if not project.active:
			raise ValidationError(
				"A consumable may only be billed to an active project. The user's project is inactive."
			)
		if not project.account.active:
			raise ValidationError(
				"A consumable may only be billed to a project that belongs to an active account. The user's account is inactive."
			)
		return project

	def clean_quantity(self):
		quantity = self.cleaned_data["quantity"]
		if quantity < 1:
			raise ValidationError("Please specify a valid quantity of items to withdraw.")
		return quantity

	def clean(self):
		if any(self.errors):
			return
		cleaned_data = super().clean()
		quantity = cleaned_data["quantity"]
		consumable = cleaned_data["consumable"]
		if quantity > consumable.quantity:
			raise ValidationError(
				'There are not enough "' + consumable.name + '". (The current quantity in stock is '
				+ str(consumable.quantity)
				+ "). Please order more as soon as possible."
			)
		customer = cleaned_data["customer"]
		project = cleaned_data["project"]
		if project not in customer.active_projects():
			raise ValidationError(
				"{} is not a member of the project {}. Users can only bill to projects they belong to.".format(
					customer, project
				)
			)
		return cleaned_data


class ReservationAbuseForm(Form):
	cancellation_horizon = IntegerField(initial=6, min_value=1)
	cancellation_penalty = IntegerField(initial=10)
	target = CharField(required=False)
	start = DateField(initial=timezone.now().replace(day=1).date())
	end = DateField(initial=timezone.now().date())

	def clean_cancellation_horizon(self):
		# Convert the cancellation horizon from hours to seconds.
		return self.cleaned_data["cancellation_horizon"] * 60 * 60

	def clean_start(self):
		start = self.cleaned_data["start"]
		return timezone.make_aware(
			datetime(year=start.year, month=start.month, day=start.day, hour=0, minute=0, second=0, microsecond=0),
			timezone.get_current_timezone(),
		)

	def clean_end(self):
		end = self.cleaned_data["end"]
		return timezone.make_aware(
			datetime(year=end.year, month=end.month, day=end.day, hour=23, minute=59, second=59, microsecond=999999),
			timezone.get_current_timezone(),
		)

	def get_target(self):
		target = self.cleaned_data["target"].split("|", 1)
		return ReservationItemType(target[0]), int(target[1])


class EmailBroadcastForm(Form):
	subject = CharField(required=False)
	color = ChoiceField(
		choices=(
			(bootstrap_primary_color("info"), "info"),
			(bootstrap_primary_color("success"), "success"),
			(bootstrap_primary_color("warning"), "warning"),
			(bootstrap_primary_color("danger"), "danger"),
		)
	)
	title = CharField(required=False)
	greeting = CharField(required=False)
	contents = CharField(required=False)
	copy_me = BooleanField(initial=True)

	audience = ChoiceField(choices=[("tool", "tool"), ("project", "project"), ("account", "account"), ("area", "area"), ("user", "user")])
	selection = CharField(required=False)
	no_type = BooleanField(initial=False, required=False)
	only_active_users = BooleanField(initial=True)

	def clean_title(self):
		return self.cleaned_data["title"].upper()

	def clean_selection(self):
		return self.data.getlist('selection')


class AlertForm(ModelForm):
	class Meta:
		model = Alert
		fields = ["title", "category", "contents", "debut_time", "expiration_time"]

	def clean_category(self):
		category = self.cleaned_data["category"]
		if not category and AlertCategory.objects.exists():
			raise ValidationError("Please select a category.")
		return category


class ScheduledOutageForm(ModelForm):
	def __init__(self, *positional_arguments, **keyword_arguments):
		super().__init__(*positional_arguments, **keyword_arguments)
		self.fields["details"].required = True

	class Meta:
		model = ScheduledOutage
		fields = ["details", "start", "end", "resource", "category"]


class UserPreferencesForm(ModelForm):
	class Meta:
		model = UserPreferences
		fields = [
			"attach_created_reservation",
			"attach_cancelled_reservation",
			"display_new_buddy_request_notification",
			"display_new_buddy_request_reply_notification",
			"email_new_buddy_request_reply",
		]


class BuddyRequestForm(ModelForm):
	class Meta:
		model = BuddyRequest
		fields = "__all__"

	def clean(self):
		if any(self.errors):
			return
		cleaned_data = super().clean()
		start = cleaned_data.get("start")
		end = cleaned_data.get("end")
		if end < start:
			self.add_error("end", "The end must be later than the start")
		return cleaned_data


class TemporaryPhysicalAccessRequestForm(ModelForm):
	class Meta:
		model = TemporaryPhysicalAccessRequest
		fields = ["start_time", "end_time", "physical_access_level", "other_users", "description"]

	def clean(self):
		if any(self.errors):
			return
		cleaned_data = super().clean()
		other_users = len(cleaned_data.get("other_users")) if "other_users" in cleaned_data else 0
		minimum_total_users = quiet_int(get_customization("access_requests_minimum_users", 2))
		if other_users < minimum_total_users -1:
			self.add_error("other_users", f"You need at least {minimum_total_users-1} other {'buddy' if minimum_total_users == 2 else 'buddies'} for this request")
		return cleaned_data


class StaffAbsenceForm(ModelForm):
	class Meta:
		model = StaffAbsence
		fields = "__all__"


def nice_errors(form, non_field_msg="General form errors"):
	result = ErrorDict()
	if isinstance(form, BaseForm):
		for field, errors in form.errors.items():
			if field == NON_FIELD_ERRORS:
				key = non_field_msg
			else:
				key = form.fields[field].label
			result[key] = errors
	return result



# Generated by Django 2.2.10 on 2020-05-17 20:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0015_area_maximum_capacity'),
    ]

    operations = [
        migrations.AddField(
            model_name='area',
            name='reservation_warning',
            field=models.PositiveIntegerField(blank=True, help_text='The number of simultaneous users (with at least one reservation in this area) allowed before a warning is displayed when creating a reservation.', null=True),
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.views.decorators.http import require_GET

from NEMO.decorators import disable_session_expiry_refresh
from NEMO.models import ReservationItemType
from NEMO.views.status_dashboard import create_tool_summary, create_area_summary


@login_required
@require_GET
@disable_session_expiry_refresh
def refresh_sidebar_icons(request, item_type=None):
	area_summary = []
	tool_summary = []
	item_type = ReservationItemType(item_type)
	if item_type == ReservationItemType.NONE:
		tool_summary = create_tool_summary()
		area_summary = create_area_summary()
	elif item_type == ReservationItemType.AREA:
		area_summary = create_area_summary()
	elif item_type == ReservationItemType.TOOL:
		tool_summary = create_tool_summary()

	return render(request, 'refresh_sidebar_icons.html', {'tool_summary':tool_summary, 'area_summary':area_summary})


from datetime import date
from typing import Optional

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseBadRequest
from django.shortcuts import get_object_or_404, redirect, render
from django.template.defaultfilters import linebreaksbr
from django.urls import reverse
from django.views.decorators.http import require_GET, require_POST, require_http_methods

from NEMO.exceptions import (
	InactiveUserError,
	NoActiveProjectsForUserError,
	NoPhysicalAccessUserError,
	PhysicalAccessExpiredUserError,
)
from NEMO.forms import BuddyRequestForm
from NEMO.models import Area, BuddyRequest, BuddyRequestMessage, User
from NEMO.views.customization import get_customization
from NEMO.views.notifications import (
	create_buddy_reply_notification,
	create_buddy_request_notification,
	delete_notification,
	get_notifications,
)
from NEMO.views.policy import check_policy_to_enter_any_area


@login_required
@require_GET
def buddy_requests(request):
	mark_requests_expired()
	buddy_requests = BuddyRequest.objects.filter(expired=False, deleted=False).order_by(
		"start", "end", "-creation_time"
	)
	# extend buddy request to add whether the current user can reply
	for buddy_request in buddy_requests:
		buddy_request.user_reply_error = check_user_reply_error(buddy_request, request.user)
	dictionary = {
		"buddy_requests": buddy_requests,
		"buddy_board_description": get_customization("buddy_board_description"),
		"request_notifications": get_notifications(request.user, BuddyRequest),
		"reply_notifications": get_notifications(request.user, BuddyRequestMessage),
	}
	return render(request, "requests/buddy_requests/buddy_requests.html", dictionary)


@login_required
@require_http_methods(["GET", "POST"])
def create_buddy_request(request, request_id=None):
	try:
		buddy_request = BuddyRequest.objects.get(id=request_id)
	except BuddyRequest.DoesNotExist:
		buddy_request = None

	dictionary = {"areas": Area.objects.filter(buddy_system_allowed=True)}

	if buddy_request:
		if buddy_request.replies.count() > 0:
			return HttpResponseBadRequest("You are not allowed to edit a request that has replies.")
		if buddy_request.user != request.user:
			return HttpResponseBadRequest("You are not allowed to edit a request you didn't create.")

	if request.method == "POST":
		form = BuddyRequestForm(request.POST, instance=buddy_request)
		form.fields["user"].required = False
		form.fields["creation_time"].required = False
		if form.is_valid():
			form.instance.user = request.user
			created_buddy_request = form.save()
			create_buddy_request_notification(created_buddy_request)
			return redirect("user_requests", "buddy")
		else:
			dictionary["form"] = form
			return render(request, "requests/buddy_requests/buddy_request.html", dictionary)
	else:
		form = BuddyRequestForm(instance=buddy_request)
		form.user = request.user
		dictionary["form"] = form
		return render(request, "requests/buddy_requests/buddy_request.html", dictionary)


@login_required
@require_GET
def delete_buddy_request(request, request_id):
	buddy_request = get_object_or_404(BuddyRequest, id=request_id)

	if buddy_request.replies.count() > 0:
		return HttpResponseBadRequest("You are not allowed to delete a request that has replies.")
	if buddy_request.user != request.user:
		return HttpResponseBadRequest("You are not allowed to delete a request you didn't create.")

	buddy_request.deleted = True
	buddy_request.save(update_fields=["deleted"])
	delete_notification(BuddyRequest, buddy_request.id)
	return redirect("user_requests", "buddy")


@login_required
@require_POST
def buddy_request_reply(request, request_id):
	buddy_request = get_object_or_404(BuddyRequest, id=request_id)
	user: User = request.user
	message_content = request.POST["reply_content"]

	error_message = check_user_reply_error(buddy_request, user)
	if error_message:
		return HttpResponseBadRequest(error_message)
	elif message_content:
		reply = BuddyRequestMessage()
		reply.buddy_request = buddy_request
		reply.content = message_content
		reply.author = user
		reply.save()
		create_buddy_reply_notification(reply)
		email_interested_parties(
			reply, request.build_absolute_uri(f"{reverse('user_requests', kwargs={'tab': 'buddy'})}?#{reply.id}")
		)
	return redirect("user_requests", "buddy")


def email_interested_parties(reply: BuddyRequestMessage, reply_url):
	creator: User = reply.buddy_request.user
	for user in reply.buddy_request.creator_and_reply_users():
		if user != reply.author and (user == creator or user.get_preferences().email_new_buddy_request_reply):
			creator_display = f"{creator.get_name()}'s" if creator != user else "your"
			creator_display_his = creator_display if creator != reply.author else "his"
			subject = f"New reply on {creator_display} buddy request"
			message = f"""{reply.author.get_name()} also replied to {creator_display_his} buddy request:
<br><br>
{linebreaksbr(reply.content)}
<br><br>
Please visit {reply_url} to reply"""
			user.email_user(subject=subject, content=message, from_email=settings.SERVER_EMAIL)


def check_user_reply_error(buddy_request: BuddyRequest, user: User) -> Optional[str]:
	error_message = None
	try:
		check_policy_to_enter_any_area(user)
	except InactiveUserError:
		error_message = "You cannot reply to this request because your account has been deactivated"
	except NoActiveProjectsForUserError:
		error_message = "You cannot reply to this request because you don't have any active projects"
	except PhysicalAccessExpiredUserError:
		error_message = "You cannot reply to this request because your facility access has expired"
	except NoPhysicalAccessUserError:
		error_message = "You cannot reply to this request because you do not have access to any areas"
	else:
		if buddy_request.area not in user.accessible_areas():
			error_message = (
				f"You cannot reply to this request because you do not have access to the {buddy_request.area.name}"
			)
	return error_message


def mark_requests_expired():
	BuddyRequest.objects.filter(expired=False, deleted=False, end__lt=date.today()).update(expired=True)



from django.db.models import F, Q
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, render
from django.views.decorators.http import require_GET, require_POST

from NEMO.decorators import staff_member_required
from NEMO.models import Project, StaffCharge, UsageEvent, User
from NEMO.utilities import (
	BasicDisplayTable,
	export_format_datetime,
	format_datetime,
	get_month_timeframe,
	month_list,
	parse_start_and_end_date,
)


@staff_member_required
@require_GET
def remote_work(request):
	if request.GET.get("start_date") and request.GET.get("end_date"):
		start_date, end_date = parse_start_and_end_date(request.GET.get("start_date"), request.GET.get("end_date"))
	else:
		start_date, end_date = get_month_timeframe()

	operator = request.GET.get("operator")
	if operator:
		if operator == "all staff":
			operator = None
		else:
			operator = get_object_or_404(User, id=operator)
	else:
		operator = request.user

	project = request.GET.get("project")
	if project and project != "all projects":
		project = get_object_or_404(Project, id=project)
	else:
		project = None
	usage_events = UsageEvent.objects.filter(
		operator__is_staff=True, start__gte=start_date, start__lte=end_date
	).exclude(operator=F("user"))
	staff_charges = StaffCharge.objects.filter(start__gte=start_date, start__lte=end_date)
	if operator:
		usage_events = usage_events.exclude(~Q(operator_id=operator.id))
		staff_charges = staff_charges.exclude(~Q(staff_member_id=operator.id))
	if project:
		usage_events = usage_events.filter(project=project)
		staff_charges = staff_charges.filter(project=project)

	csv_export = bool(request.GET.get("csv", False))
	if csv_export:
		table_result = BasicDisplayTable()
		TYPE, ID, ITEM, STAFF, CUSTOMER, PROJECT, START, END = (
			"item_type",
			"item_id",
			"item",
			"staff_member",
			"customer",
			"project",
			"start_date",
			"end_date",
		)
		table_result.headers = [
			(TYPE, "Item Type"),
			(ID, "Item Id"),
			(ITEM, "Item"),
			(STAFF, "Staff"),
			(CUSTOMER, "Customer"),
			(PROJECT, "Project"),
			(START, "Start"),
			(END, "End"),
		]
		for usage in usage_events:
			table_result.add_row(
				{
					ID: usage.tool.id,
					TYPE: "Tool Usage",
					ITEM: usage.tool,
					STAFF: usage.operator,
					CUSTOMER: usage.user,
					START: format_datetime(usage.start, "SHORT_DATETIME_FORMAT"),
					END: format_datetime(usage.end, "SHORT_DATETIME_FORMAT") if usage.end else "",
					PROJECT: usage.project,
				}
			)
		for staff_charge in staff_charges:
			for access in staff_charge.areaaccessrecord_set.all():
				table_result.add_row(
					{
						ID: access.area.id,
						TYPE: "Area Access",
						ITEM: access.area,
						STAFF: staff_charge.staff_member,
						CUSTOMER: access.customer,
						START: format_datetime(access.start, "SHORT_DATETIME_FORMAT"),
						END: format_datetime(access.end, "SHORT_DATETIME_FORMAT") if access.end else "",
						PROJECT: access.project,
					}
				)
			table_result.add_row(
				{
					ID: staff_charge.id,
					TYPE: "Staff Charge",
					ITEM: "Staff Charge",
					STAFF: staff_charge.staff_member,
					CUSTOMER: staff_charge.customer,
					START: format_datetime(staff_charge.start, "SHORT_DATETIME_FORMAT"),
					END: format_datetime(staff_charge.end, "SHORT_DATETIME_FORMAT") if staff_charge.end else "",
					PROJECT: staff_charge.project,
				}
			)
		response = table_result.to_csv()
		filename = f"remote_work_{export_format_datetime(start_date, t_format=False)}_to_{export_format_datetime(end_date, t_format=False)}.csv"
		response["Content-Disposition"] = f'attachment; filename="{filename}"'
		return response
	dictionary = {
		"usage": usage_events,
		"staff_charges": staff_charges,
		"staff_list": User.objects.filter(is_staff=True),
		"project_list": Project.objects.filter(active=True),
		"start_date": start_date,
		"end_date": end_date,
		"month_list": month_list(),
		"selected_staff": operator.id if operator else "all staff",
		"selected_project": project.id if project else "all projects",
	}
	return render(request, "remote_work.html", dictionary)


@staff_member_required
@require_POST
def validate_staff_charge(request, staff_charge_id):
	staff_charge = get_object_or_404(StaffCharge, id=staff_charge_id)
	staff_charge.validated = True
	staff_charge.save()
	return HttpResponse()


@staff_member_required
@require_POST
def validate_usage_event(request, usage_event_id):
	usage_event = get_object_or_404(UsageEvent, id=usage_event_id)
	usage_event.validated = True
	usage_event.save()
	return HttpResponse()



from rest_framework import permissions


class BillingAPI(permissions.BasePermission):
	""" Checks that a user has permission to use the NEMO RESTful API for billing purposes. """
	def has_permission(self, request, view):
		if request and request.user.has_perm('NEMO.use_billing_api'):
				return True
		return False



from django.conf.urls import url
from django.urls import path, include

from NEMO.apps.kiosk import views
from NEMO.views import area_access

urlpatterns = [
	# Tablet kiosk
	path('kiosk/', include([
		url(r'^occupancy/$', area_access.occupancy, name='kiosk_occupancy'),
		url(r'^enable_tool/$', views.enable_tool, name='enable_tool_from_kiosk'),
		url(r'^disable_tool/$', views.disable_tool, name='disable_tool_from_kiosk'),
		url(r'^reserve_tool/$', views.reserve_tool, name='reserve_tool_from_kiosk'),
		url(r'^cancel_reservation/(?P<reservation_id>\d+)/$', views.cancel_reservation,	name='cancel_reservation_from_kiosk'),
		url(r'^choices/$', views.choices, name='kiosk_choices'),
		url(r'^category_choices/(?P<category>.+)/(?P<user_id>\d+)/$', views.category_choices, name='kiosk_category_choices'),
		url(r'^tool_information/(?P<tool_id>\d+)/(?P<user_id>\d+)/(?P<back>back_to_start|back_to_category)/$', views.tool_information, name='kiosk_tool_information'),
		url(r'^tool_reservation/(?P<tool_id>\d+)/(?P<user_id>\d+)/(?P<back>back_to_start|back_to_category)/$', views.tool_reservation, name='kiosk_tool_reservation'),
		url(r'^(?P<location>.+)/$', views.kiosk, name='kiosk'),
		url(r'^$', views.kiosk, name='kiosk'),
	]))
]



# Generated by Django 2.2.13 on 2020-11-24 15:52

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0024_contactinformation_user'),
    ]

    operations = [
        migrations.CreateModel(
            name='ToolUsageCounter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of this counter', max_length=200)),
                ('description', models.TextField(blank=True, help_text='The counter description to be displayed next to it on the tool control page', null=True)),
                ('value', models.PositiveIntegerField(default=0, help_text='The current value of this counter')),
                ('tool_usage_question', models.CharField(help_text="The name of the tool's post usage question which should be used to increment this counter", max_length=200)),
                ('last_reset_value', models.PositiveIntegerField(blank=True, help_text='The last value before the counter was reset', null=True)),
                ('last_reset', models.DateTimeField(blank=True, help_text='The date and time this counter was last reset', null=True)),
                ('is_active', models.BooleanField(default=True, help_text='The state of the counter')),
                ('last_reset_by', models.ForeignKey(blank=True, help_text='The user who last reset this counter', null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
                ('tool', models.ForeignKey(help_text='The tool this counter is for.', on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool')),
            ],
        ),
        migrations.CreateModel(
            name='PhysicalAccessException',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of this exception that will be displayed as the policy problem', max_length=100)),
                ('start_time', models.DateTimeField(help_text='The start of the exception, after which users will be denied access.')),
                ('end_time', models.DateTimeField(help_text='The end of the exception, after which users will be allowed access again')),
                ('physical_access_levels', models.ManyToManyField(blank=True, to='NEMO.PhysicalAccessLevel')),
            ],
            options={
                'ordering': ['-start_time'],
            },
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.http import HttpResponseNotFound
from django.shortcuts import get_object_or_404, render
from django.views.decorators.http import require_GET

from NEMO.models import AreaAccessRecord, Reservation, ScheduledOutage, UsageEvent
from NEMO.utilities import format_datetime


@login_required
@require_GET
def reservation_details(request, reservation_id):
	reservation = get_object_or_404(Reservation, id=reservation_id)
	popup_view = request.GET.get("popup_view")
	if reservation.cancelled:
		error_message = "This reservation was cancelled by {0} at {1}.".format(
			reservation.cancelled_by, format_datetime(reservation.cancellation_time)
		)
		return HttpResponseNotFound(error_message)
	reservation_project_can_be_changed = (request.user.is_staff or request.user == reservation.user) and reservation.has_not_ended and reservation.has_not_started and reservation.user.active_project_count() > 1

	template_data = {
		"reservation": reservation,
		"reservation_project_can_be_changed": reservation_project_can_be_changed,
		"popup_view": popup_view,
	}

	return render(request, "event_details/reservation_details.html", template_data)


@login_required
@require_GET
def outage_details(request, outage_id):
	outage = get_object_or_404(ScheduledOutage, id=outage_id)
	return render(
		request, "event_details/outage_details.html", {"outage": outage, "popup_view": request.GET.get("popup_view")}
	)


@login_required
@require_GET
def usage_details(request, event_id):
	event = get_object_or_404(UsageEvent, id=event_id)
	return render(request, "event_details/usage_details.html", {"event": event, "popup_view": request.GET.get("popup_view")})


@login_required
@require_GET
def area_access_details(request, event_id):
	event = get_object_or_404(AreaAccessRecord, id=event_id)
	return render(
		request, "event_details/area_access_details.html", {"event": event, "popup_view": request.GET.get("popup_view")}
	)



import importlib
import json
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Dict, List, Union

from django.conf import settings

from NEMO.models import Consumable, Tool

rates_logger = getLogger(__name__)

class Rates(ABC):

	rates = None

	@abstractmethod
	def load_rates(self):
		pass

	def get_consumable_rates(self, consumables: List[Consumable]) -> Dict[str, str]:
		if self.rates:
			return {consumable.name : self.get_consumable_rate(consumable) for consumable in consumables}

	@abstractmethod
	def get_consumable_rate(self, consumable: Consumable) -> str:
		pass

	def get_tool_rates(self, tools: List[Tool]) -> Dict[str, str]:
		if self.rates:
			return {tool.name : self.get_tool_rate(tool) for tool in tools}

	@abstractmethod
	def get_tool_rate(self, tool: Tool) -> str:
		pass


class NISTRates(Rates):

	consumable_rate_class = 'inventory_rate'
	tool_rate_class = 'primetime_eq_hourly_rate'
	tool_training_rate_class = 'training_individual_hourly_rate'
	tool_training_group_rate_class = 'training_group_hourly_rate'

	full_cost_rate_class = 'full cost'
	shared_cost_rate_class = 'cost shared'

	def load_rates(self, force_reload=False):
		if force_reload:
			self.rates = None
		if not self.rates:
			json_data = None
			try:
				rates_file = getattr(settings, 'RATES_FILE', settings.MEDIA_ROOT + '/rates.json')
				json_data = open(rates_file)
				self.rates = json.load(json_data)
				rates_logger.info("found rates file and loaded rates")
			except FileNotFoundError as e:
				if hasattr(settings, 'RATES_FILE'):
					rates_logger.exception(e)
				else:
					rates_logger.debug("no rates file, skipping loading rates")
			except Exception as e:
				rates_logger.error("error loading rates")
				rates_logger.exception(e)
			finally:
				if json_data:
					json_data.close()

	def get_consumable_rate(self, consumable) -> str:
		full_cost_rate = self._get_rate_by_table_id_and_class(consumable, self.consumable_rate_class, self.full_cost_rate_class)
		if full_cost_rate:
			return "Cost <b>${:0,.2f}</b>".format(full_cost_rate)

	def get_tool_rate(self, tool: Tool) -> str:
		full_cost_rate = self._get_rate_by_table_id_and_class(tool, self.tool_rate_class, self.full_cost_rate_class)
		shared_cost_rate = self._get_rate_by_table_id_and_class(tool, self.tool_rate_class, self.shared_cost_rate_class)
		if full_cost_rate or shared_cost_rate:
			result = "Tool rates:"
			if tool.is_parent_tool():
				result += "<br> " + tool.name + ":"
			if full_cost_rate:
				result += " Full Cost <b>${:0,.2f}</b>".format(full_cost_rate)
			if shared_cost_rate:
				result += " Shared Cost <b>${:0,.2f}</b>".format(shared_cost_rate)
			if tool.is_parent_tool():
				for child_tool in tool.tool_children_set.all():
					child_full_cost_rate = self._get_rate_by_table_id_and_class(child_tool, self.tool_rate_class, self.full_cost_rate_class)
					child_shared_cost_rate = self._get_rate_by_table_id_and_class(child_tool, self.tool_rate_class, self.shared_cost_rate_class)
					if child_full_cost_rate or child_shared_cost_rate:
						result += "<br> " + child_tool.name + ":"
						if child_full_cost_rate:
							result += " Full Cost <b>${:0,.2f}</b>".format(child_full_cost_rate)
						if child_shared_cost_rate:
							result += " Shared Cost <b>${:0,.2f}</b>".format(child_shared_cost_rate)
			training_rate = self._get_rate_by_table_id_and_class(tool, self.tool_training_rate_class, self.full_cost_rate_class)
			training_group_rate = self._get_rate_by_table_id_and_class(tool, self.tool_training_group_rate_class, self.full_cost_rate_class)
			if training_rate or training_group_rate:
				result += "<br>Training rates:"
				if training_rate:
					result += " Individual <b>${:0,.2f}</b>".format(training_rate)
				if training_group_rate:
					result += " Group <b>${:0,.2f}</b>".format(training_group_rate)
			return result


	def _get_rate_by_table_id_and_class(self, item: Union[Consumable, Tool], table_id, rate_claz) -> float:
		if self.rates:
			matching_rates = list(filter(lambda rate: rate['table_id'] == table_id and rate['rate_class'] == rate_claz and rate['item_id'] == item.id, self.rates))
			if matching_rates:
				return matching_rates[0]['rate']


def get_rate_class():
	rates_class = getattr(settings, "RATES_CLASS", "NEMO.rates.NISTRates")
	assert isinstance(rates_class, str)
	pkg, attr = rates_class.rsplit(".", 1)
	ret = getattr(importlib.import_module(pkg), attr)
	return ret()


rate_class = get_rate_class()


from django.db import migrations

from NEMO.migrations_utils import create_news_for_version


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0033_version_3_12_0'),
    ]

    def new_version_news(apps, schema_editor):
        create_news_for_version(apps, "3.13.0")

    operations = [
        migrations.RunPython(new_version_news),
    ]



from NEMO.models import Area, Notification, PhysicalAccessLevel, Tool, User
from NEMO.views.customization import get_customization
from NEMO.views.notifications import get_notification_counts


def show_logout_button(request):
	return {"logout_allowed": True}


def hide_logout_button(request):
	return {"logout_allowed": False}


def base_context(request):
	try:
		if "no_header" in request.GET:
			if request.GET["no_header"] == "True":
				request.session["no_header"] = True
			else:
				request.session["no_header"] = False
	except:
		request.session["no_header"] = False
	try:
		facility_name = get_customization("facility_name")
	except:
		facility_name = "Facility"
	try:
		site_title = get_customization("site_title")
	except:
		site_title = ""
	try:
		tools_exist = Tool.objects.filter(visible=True).exists()
	except:
		tools_exist = False
	try:
		areas_exist = Area.objects.exists() and PhysicalAccessLevel.objects.exists()
	except:
		areas_exist = False
	try:
		buddy_system_areas_exist = Area.objects.filter(buddy_system_allowed=True).exists()
	except:
		buddy_system_areas_exist = False
	try:
		access_user_request_allowed_exist = PhysicalAccessLevel.objects.filter(allow_user_request=True).exists()
	except:
		access_user_request_allowed_exist = False
	try:
		notification_counts = get_notification_counts(request.user)
	except:
		notification_counts = {}
	try:
		buddy_notification_count = notification_counts.get(Notification.Types.BUDDY_REQUEST, 0)
		buddy_notification_count += notification_counts.get(Notification.Types.BUDDY_REQUEST_REPLY, 0)
	except:
		buddy_notification_count = 0
	try:
		temporary_access_notification_count = notification_counts.get(Notification.Types.TEMPORARY_ACCESS_REQUEST, 0)
	except:
		temporary_access_notification_count = 0
	try:
		facility_managers_exist = User.objects.filter(is_active=True, is_facility_manager=True).exists()
	except:
		facility_managers_exist = False
	return {
		"facility_name": facility_name,
		"site_title": site_title,
		"device": request.device,
		"tools_exist": tools_exist,
		"areas_exist": areas_exist,
		"buddy_system_areas_exist": buddy_system_areas_exist,
		"access_user_request_allowed_exist": access_user_request_allowed_exist,
		"notification_counts": notification_counts,
		"buddy_notification_count": buddy_notification_count,
		"temporary_access_notification_count": temporary_access_notification_count,
		"facility_managers_exist": facility_managers_exist,
		"no_header": request.session.get("no_header", False),
	}



from django.db import migrations, models

from NEMO.migrations_utils import create_news_for_version


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0031_version_3_10_0'),
    ]

    def new_version_news(apps, schema_editor):
        create_news_for_version(apps, "3.11.0")

    operations = [
        migrations.RunPython(new_version_news),
        migrations.CreateModel(
            name='ReservationQuestions',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of this ', max_length=100)),
                ('questions', models.TextField(help_text='Upon making a reservation, the user will be asked these questions. This field will only accept JSON format')),
                ('tool_reservations', models.BooleanField(default=True, help_text='Check this box to apply these questions to tool reservations')),
                ('only_for_tools', models.ManyToManyField(blank=True, help_text='Select the tools these questions only apply to. Leave blank for all tools', to='NEMO.Tool')),
                ('area_reservations', models.BooleanField(default=False, help_text='Check this box to apply these questions to area reservations')),
                ('only_for_areas', models.ManyToManyField(blank=True, help_text='Select the areas these questions only apply to. Leave blank for all areas', to='NEMO.Area')),
                ('only_for_projects', models.ManyToManyField(blank=True, help_text='Select the projects these questions only apply to. Leave blank for all projects', to='NEMO.Project')),
            ],
            options={
                'verbose_name_plural': 'Reservation questions',
                'ordering': ['name'],
            },
        ),
        migrations.AddField(
            model_name='reservation',
            name='question_data',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='news',
            name='pinned',
            field=models.BooleanField(default=False, help_text='Check this box to keep this story at the top of the news feed'),
        ),
        migrations.CreateModel(
            name='AccountType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.AddField(
            model_name='account',
            name='start_date',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='project',
            name='start_date',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='account',
            name='type',
            field=models.ForeignKey(blank=True, null=True, on_delete=models.deletion.SET_NULL, to='NEMO.AccountType'),
        ),
    ]



# Generated by Django 2.2.10 on 2020-04-09 19:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0013_version_2_1_0'),
    ]

    def create_proxr_interlock_card_category(apps, schema_editor):
        InterlockCardCategory = apps.get_model("NEMO", "InterlockCardCategory")
        if not InterlockCardCategory.objects.filter(key="proxr").exists():
            proxr_category = InterlockCardCategory.objects.create(name="ProXr", key="proxr")
            proxr_category.save()

    operations = [
        migrations.AddField(
            model_name='comment',
            name='staff_only',
            field=models.BooleanField(default=False),
        ),
        migrations.RunPython(create_proxr_interlock_card_category),
    ]



# Generated by Django 2.2.13 on 2020-06-30 19:38

import datetime
import NEMO.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0020_version_3_0_0'),
    ]

    operations = [
        migrations.AddField(
            model_name='area',
            name='abuse_email',
            field=NEMO.fields.MultiEmailField(blank=True, help_text='An email will be sent to this address when users overstay in the area or in children areas (logged in with expired reservation). A comma-separated list can be used.', max_length=2000, null=True),
        ),
        migrations.AddField(
            model_name='area',
            name='reservation_email',
            field=NEMO.fields.MultiEmailField(blank=True, help_text='An email will be sent to this address when users create or cancel reservations in the area or in children areas. A comma-separated list can be used.', max_length=2000, null=True),
        ),
        migrations.AlterField(
            model_name='physicalaccesslevel',
            name='schedule',
            field=models.IntegerField(choices=[(0, 'Always'), (1, 'Weekdays'), (2, 'Weekends')]),
        ),
        migrations.AddField(
            model_name='physicalaccesslevel',
            name='weekdays_end_time',
            field=models.TimeField(blank=True, default=datetime.time(0, 0), help_text='The weekday access end time', null=True),
        ),
        migrations.AddField(
            model_name='physicalaccesslevel',
            name='weekdays_start_time',
            field=models.TimeField(blank=True, default=datetime.time(7, 0), help_text='The weekday access start time', null=True),
        ),
    ]



from distutils.util import strtobool

from django import forms
from django.core import validators
from django.core.exceptions import ValidationError
from django.core.validators import EmailValidator
from django.db import models
from django.db.models import Field, TextField
from django.db.models.lookups import BuiltinLookup
from django.utils.translation import gettext_lazy as _

DEFAULT_SEPARATOR = ","


@Field.register_lookup
class IsEmpty(BuiltinLookup):
	# Custom lookup allowing to use __isempty in filters and REST API
	lookup_name = "isempty"
	prepare_rhs = False

	def as_sql(self, compiler, connection):
		sql, params = compiler.compile(self.lhs)
		condition = self.rhs if isinstance(self.rhs, bool) else bool(strtobool(self.rhs))
		if condition:
			return "%s IS NULL or %s = ''" % (sql, sql), params
		else:
			if getattr(connection, "vendor", "") == "oracle" and isinstance(self.lhs.field, TextField):
				# we need to handle textfields for oracle differently as they are set as clobs
				return "dbms_lob.getlength(%s) != 0" % sql, params
			else:
				return "%s <> ''" % sql, params


class MultiEmailWidget(forms.TextInput):
	is_hidden = False
	separator = DEFAULT_SEPARATOR

	def __init__(self, attrs=None):
		super().__init__(attrs={"size": "129", **(attrs or {})})

	def prep_value(self, value):
		if self.attrs and "separator" in self.attrs:
			self.separator = self.attrs.pop("separator")

		if value in validators.EMPTY_VALUES + ("[]",):
			return ""
		elif isinstance(value, str):
			return value
		elif isinstance(value, list):
			return self.separator.join(value)
		raise ValidationError("Invalid format.")

	def render(self, name, value, **kwargs):
		value = self.prep_value(value)
		return super().render(name, value, **kwargs)


class MultiEmailFormField(forms.CharField):
	widget = MultiEmailWidget

	def __init__(self, separator=DEFAULT_SEPARATOR, **kwargs):
		self.separator = separator
		super().__init__(**kwargs)

	def widget_attrs(self, widget):
		attrs = super().widget_attrs(widget)
		attrs["separator"] = self.separator
		return attrs

	def prepare_value(self, value):
		if type(value) is list:
			return self.separator.join(value)
		else:
			return value


class MultiEmailField(models.CharField):
	description = "A multi e-mail field stored as a configurable character separated string"

	def __init__(self, separator=DEFAULT_SEPARATOR, *args, **kwargs):
		self.email_validator = EmailValidator(
			message=_("Enter a valid email address or a list separated by {}").format(separator)
		)
		self.separator = separator
		kwargs.setdefault("max_length", 2000)
		super().__init__(*args, **kwargs)

	def deconstruct(self):
		name, path, args, kwargs = super().deconstruct()
		# Only include kwarg if it's not the default
		if self.separator != DEFAULT_SEPARATOR:
			kwargs["separator"] = self.separator
		return name, path, args, kwargs

	def formfield(self, **kwargs):
		# We are forcing our form class and widget here, replacing potential overrides from kwargs
		return super().formfield(
			**{**kwargs, "form_class": MultiEmailFormField, "widget": MultiEmailWidget, "separator": self.separator}
		)

	def validate(self, value, model_instance):
		"""Check if value consists only of valid emails."""
		super().validate(self.get_prep_value(value), model_instance)
		for email in value:
			self.email_validator(email.strip())

	def from_db_value(self, value, expression, connection):
		return self.to_python(value)

	def get_prep_value(self, value):
		if value is None:
			return value
		return self.separator.join(value)

	def to_python(self, value):
		if isinstance(value, list):
			return value
		if value is None:
			return []
		return [address.strip() for address in value.split(self.separator)]

	def value_to_string(self, obj):
		value = self.value_from_object(obj)
		return self.get_prep_value(value)



from datetime import datetime, timedelta
from typing import Dict, List, Optional, Union

from dateutil.relativedelta import relativedelta
from dateutil.rrule import DAILY, rrule
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.db.models import F, Prefetch, Q, QuerySet
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import timezone
from django.views.decorators.http import require_GET, require_http_methods

from NEMO.decorators import disable_session_expiry_refresh, facility_manager_required
from NEMO.forms import StaffAbsenceForm
from NEMO.model_tree import ModelTreeHelper, TreeItem, get_area_model_tree
from NEMO.models import (
	Area,
	AreaAccessRecord,
	ClosureTime,
	Resource,
	ScheduledOutage,
	StaffAbsence,
	StaffAvailability,
	Task,
	Tool,
	UsageEvent,
	User,
)
from NEMO.utilities import (
	BasicDisplayTable,
	as_timezone,
	beginning_of_the_day,
	export_format_datetime,
	format_datetime,
	quiet_int,
)
from NEMO.views.customization import get_customization


@login_required
@require_GET
@disable_session_expiry_refresh
def status_dashboard(request, tab=None):
	"""
	Present a web page to allow users to view the status and usage of all tools.
	"""
	interest = request.GET.get("interest")
	if interest is None:
		csv_export = bool(request.GET.get("csv", False))
		if tab == "staff" and csv_export:
			return get_staff_status(request, csv_export)
		dictionary = {
			"tab": tab if tab else "occupancy",
			"show_staff_status": show_staff_status(request),
			**get_tools_dictionary(),
			**get_occupancy_dictionary(request),
			**get_staff_status(request),
		}
		return render(request, "status_dashboard/status_dashboard.html", dictionary)
	elif interest == "tools":
		return render(request, "status_dashboard/tools.html", get_tools_dictionary())
	elif interest == "occupancy":
		return render(request, "status_dashboard/occupancy.html", get_occupancy_dictionary(request))


def get_tools_dictionary():
	return {"tool_summary": create_tool_summary()}


def get_occupancy_dictionary(request):
	area_items, no_occupants = process_area_access_record_with_parents(request.user)
	return {"area_items": area_items, "no_occupants": no_occupants}


def get_staff_status(request, csv_export=False) -> Union[Dict, HttpResponse]:
	# Timestamp allows us to know which week/month to show. Defaults to current week
	# Everything here is dealing with date/times without timezones to avoid issues with DST etc
	view = request.GET.get("view", "week")
	timestamp = quiet_int(request.GET.get("timestamp", datetime.now().timestamp()), datetime.now().timestamp())
	today = beginning_of_the_day(datetime.fromtimestamp(timestamp), in_local_timezone=False)
	weekdays_only = get_customization("dashboard_staff_status_weekdays_only")
	first_day = (
		today.isoweekday()
		if not weekdays_only and get_customization("dashboard_staff_status_first_day_of_week") == "0"
		else today.weekday()
	)
	start = today.replace(day=1) if view == "month" else today - timedelta(days=first_day)
	end_delta = relativedelta(months=1) if view == "month" else timedelta(weeks=1)
	end = start + end_delta - timedelta(days=1)
	# If we are only showing weekdays, we have to subtract 2 days from the end of the week. Only applies to week view
	if view == "week" and weekdays_only:
		end = end - timedelta(days=2)
	# Reset timestamp to be right in the middle of the period
	timestamp = int((start + timedelta(days=(end - start).days/2)).timestamp())
	staffs = StaffAvailability.objects.all()
	staffs.query.add_ordering(F("category").asc(nulls_last=True))
	staffs.query.add_ordering(F("staff_member__first_name").asc())
	days = rrule(DAILY, dtstart=start, until=end)
	staff_date_format = get_customization("dashboard_staff_status_date_format")
	if csv_export:
		return export_staff_status(request, staffs, days, start, end, staff_date_format)
	return {
		"staff_date_format": staff_date_format,
		"staff_absences": staff_absences_dict(staffs, days, start, end),
		"closure_times": closures_dict(days, start, end),
		"staffs": staffs,
		"days": days,
		"days_length": (end - start).days + 1,
		"page_timestamp": timestamp,
		"page_view": view,
		# Using end delta here (=/- 1 week or 1 month) to set previous and next
		"prev": int((start - end_delta).timestamp()),
		"next": int((end + end_delta).timestamp()),
	}


@facility_manager_required
@require_http_methods(["GET", "POST"])
def create_staff_absence(request, absence_id=None):
	try:
		absence = StaffAbsence.objects.get(pk=absence_id)
	except (StaffAbsence.DoesNotExist, ValueError):
		absence = StaffAbsence()
		# Set the staff if we were given its id
		try:
			staff_id = request.GET.get("staff_id", None)
			if staff_id:
				absence.staff_member = StaffAvailability.objects.get(pk=staff_id)
		except (StaffAvailability.DoesNotExist, ValueError):
			pass
	form = StaffAbsenceForm(request.POST or None, instance=absence)
	timestamp = request.GET.get("timestamp", "")
	view = request.GET.get("view", "")
	if request.POST and form.is_valid():
		form.save()
		return HttpResponse()
	dictionary = {"form": form, "staff_members": StaffAvailability.objects.all(), "page_timestamp": timestamp, "page_view": view}
	return render(request, "status_dashboard/staff_absence.html", dictionary)


@facility_manager_required
@require_GET
def delete_staff_absence(request, absence_id):
	absence = get_object_or_404(StaffAbsence, id=absence_id)
	absence.delete()
	timestamp = request.GET.get("timestamp", "")
	view = request.GET.get("view", "")
	url = reverse("status_dashboard_tab", args=["staff"])
	return redirect(f"{url}?timestamp={timestamp}&view={view}")


def staff_absences_dict(staffs, days, start, end):
	dictionary = {staff.id: {} for staff in staffs}
	absences = StaffAbsence.objects.filter(start_date__lte=end, end_date__gte=start).order_by("creation_time")
	for staff_absence in absences:
		for day in days:
			# comparing dates here so no timezone issues (dates don't have timezones)
			if staff_absence.start_date <= day.date() <= staff_absence.end_date:
				dictionary[staff_absence.staff_member.id][day.day] = staff_absence
	return dictionary


def closures_dict(days, start, end):
	dictionary = {}
	closure_times = ClosureTime.objects.filter(start_time__lte=end, end_time__gte=start)
	for closure_time in closure_times:
		for day in days:
			if as_timezone(closure_time.start_time).date() <= day.date() <= as_timezone(closure_time.end_time).date():
				dictionary[day.day] = closure_time
	return dictionary


@facility_manager_required
def export_staff_status(request, staffs, days, start, end, staff_date_format) -> HttpResponse:
	table_result = BasicDisplayTable()
	table_result.add_header(("staff", ""))
	# Add headers for each day
	for day in days:
		table_result.add_header((day.day, format_datetime(day, staff_date_format, as_current_timezone=False)))
	staff_absences = staff_absences_dict(staffs, days, start, end)
	closure_times = closures_dict(days, start, end)
	category = ""
	for staff in staffs:
		if staff.category != category:
			category = staff.category or "Other"
			# Add a whole row for the category
			table_result.add_row({"staff": category})
		# Add a row for each staff member
		staff_row = {"staff": staff.staff_member.get_name()}
		for day in days:
			staff_row[day.day] = staff.weekly_availability("In", "Out").get(day.weekday())
		for staff_id, absence_dict in staff_absences.items():
			if staff.id == staff_id:
				for day_index, absence in absence_dict.items():
					staff_row[day_index] = absence.absence_type.name + f" {absence.description or ''}"
		for day_index, closure_time in closure_times.items():
			staff_row[day_index] = f"Closed: {closure_time.closure.name}"
		table_result.add_row(staff_row)
	response = table_result.to_csv()
	filename = f"staff_status_{export_format_datetime(start, t_format=False)}_to_{export_format_datetime(end, t_format=False)}.csv"
	response["Content-Disposition"] = f'attachment; filename="{filename}"'
	return response


def show_staff_status(request):
	if not settings.ALLOW_CONDITIONAL_URLS:
		return False
	dashboard_staff_status_staff_only = get_customization("dashboard_staff_status_staff_only")
	return StaffAvailability.objects.exists() and (not dashboard_staff_status_staff_only or request.user.is_staff)


def process_area_access_record_with_parents(user: User):
	show_not_qualified_areas = get_customization("dashboard_display_not_qualified_areas")
	records = AreaAccessRecord.objects.filter(end=None, staff_charge=None)
	if not user.is_staff and show_not_qualified_areas != "enabled":
		records = records.filter(area__in=user.accessible_areas())
	records = records.prefetch_related("customer", "project", "area")
	no_occupants = not records.exists()
	area_items = None
	area_model_tree = get_area_model_tree()
	if not no_occupants:
		areas_and_parents = area_model_tree.get_ancestor_areas(
			area_model_tree.get_areas([record.area.id for record in records]), include_self=True
		)
		# Sort to have area without children before others
		areas_and_parents.sort(key=lambda x: f"{x.tree_category}zz" if x.is_leaf else f"{x.tree_category}/aa")
		area_summary = create_area_summary(
			area_tree=area_model_tree, add_resources=False, add_outages=False, add_occupants=True
		)
		area_summary_dict = {area["id"]: area for area in area_summary}
		for area_item in areas_and_parents:
			area_item.item = area_summary_dict[area_item.id]
		area_items = area_tree_helper(areas_and_parents, records)
	return area_items, no_occupants


def area_tree_helper(filtered_area: List[TreeItem], records: QuerySet, areas: Optional[List[TreeItem]] = None):
	""" Recursively build a list of areas. The resulting list is meant to be iterated over in a view """
	if areas is None:
		# Get the root areas
		areas = [area for area in filtered_area if area.is_root]
	else:
		yield "in"

	for area in areas:
		yield area
		children = [child for child in area.children if child in filtered_area]
		if len(children):
			area.leaf = False
			for x in area_tree_helper(filtered_area, records, children):
				yield x
		else:
			area.occupants = records.filter(area__id=area.id)
			area.leaf = True
	yield "out"


def create_tool_summary():
	tools = Tool.objects.filter(visible=True).prefetch_related(
		Prefetch("_requires_area_access", queryset=Area.objects.all().only("name"))
	)
	tasks = Task.objects.filter(cancelled=False, resolved=False, tool__visible=True).prefetch_related("tool")
	unavailable_resources = Resource.objects.filter(available=False).prefetch_related(
		"fully_dependent_tools", "partially_dependent_tools"
	)
	# also check for visibility on the parent if there is one (alternate tool are hidden)
	usage_events = UsageEvent.objects.filter(
		Q(end=None, tool__visible=True) | Q(end=None, tool__parent_tool__visible=True)
	).prefetch_related("operator", "user", "tool")
	scheduled_outages = ScheduledOutage.objects.filter(
		start__lte=timezone.now(), end__gt=timezone.now(), area__isnull=True
	)
	tool_summary = merge(tools, tasks, unavailable_resources, usage_events, scheduled_outages)
	tool_summary = list(tool_summary.values())
	tool_summary.sort(key=lambda x: x["name"])
	return tool_summary


def create_area_summary(area_tree: ModelTreeHelper = None, add_resources=True, add_occupants=True, add_outages=True):
	if area_tree is None:
		area_tree = get_area_model_tree()
	area_items = area_tree.items.values()
	result = {}
	for area in area_items:
		result[area.id] = {
			"name": area.name,
			"id": area.id,
			"maximum_capacity": area.maximum_capacity,
			"warning_capacity": area.item.warning_capacity(),
			"danger_capacity": area.item.danger_capacity(),
			"count_staff_in_occupancy": area.count_staff_in_occupancy,
			"count_service_personnel_in_occupancy": area.count_service_personnel_in_occupancy,
			"occupancy_count": 0,
			"occupancy": 0,
			"occupancy_staff": 0,
			"occupancy_service_personnel": 0,
			"occupants": "",
			"required_resource_is_unavailable": False,
			"scheduled_outage": False,
		}

	if add_resources:
		unavailable_resources = Resource.objects.filter(available=False).prefetch_related(
			Prefetch("dependent_areas", queryset=Area.objects.only("id"))
		)
		for resource in unavailable_resources:
			for area in resource.dependent_areas.all():
				if area.id in result:
					result[area.id]["required_resource_is_unavailable"] = True
	if add_outages:
		scheduled_outages = ScheduledOutage.objects.filter(
			start__lte=timezone.now(), end__gt=timezone.now(), tool__isnull=True
		).only("area_id", "resource_id")
		for outage in scheduled_outages:
			if outage.area_id:
				result[outage.area_id]["scheduled_outage"] = True
			elif outage.resource_id:
				for t in outage.resource.dependent_areas.values_list("id", flat=True):
					result[t]["scheduled_outage"] = True

	if add_occupants:
		occupants: List[AreaAccessRecord] = AreaAccessRecord.objects.filter(
			end=None, staff_charge=None
		).prefetch_related(
			Prefetch("customer", queryset=User.objects.all().only("first_name", "last_name", "username", "is_staff"))
		)
		for occupant in occupants:
			# Get ids for area and all the parents (so we can add occupants info on parents)
			area_ids = area_tree.get_area(occupant.area_id).ancestor_ids(include_self=True)
			if occupant.customer.is_staff:
				customer_display = f'<span class="success-highlight">{str(occupant.customer)}</span>'
			elif occupant.customer.is_service_personnel:
				customer_display = f'<span class="warning-highlight">{str(occupant.customer)}</span>'
			elif occupant.customer.is_logged_in_area_without_reservation():
				customer_display = f'<span class="danger-highlight">{str(occupant.customer)}</span>'
			else:
				customer_display = str(occupant.customer)
			for area_id in area_ids:
				if area_id in result:
					result[area_id]["occupancy"] += 1
					if occupant.customer.is_staff:
						result[area_id]["occupancy_staff"] += 1
					if occupant.customer.is_service_personnel:
						result[area_id]["occupancy_service_personnel"] += 1
					if (not occupant.customer.is_staff or result[area_id]["count_staff_in_occupancy"]) and (
							not occupant.customer.is_service_personnel
							or result[area_id]["count_service_personnel_in_occupancy"]
					):
						result[area_id]["occupancy_count"] += 1
					result[area_id]["occupants"] += (
						customer_display if not result[area_id]["occupants"] else f"<br>{customer_display}"
					)
	area_summary = list(result.values())
	area_summary.sort(key=lambda x: x["name"])
	return area_summary


def merge(tools, tasks, unavailable_resources, usage_events, scheduled_outages):
	result = {}
	tools_with_delayed_logoff_in_effect = [
		x.tool.tool_or_parent_id() for x in UsageEvent.objects.filter(end__gt=timezone.now())
	]
	parent_ids = Tool.objects.filter(parent_tool__isnull=False).values_list("parent_tool_id", flat=True)
	for tool in tools:
		result[tool.tool_or_parent_id()] = {
			"name": tool.name_or_child_in_use_name(parent_ids=parent_ids),
			"id": tool.id,
			"user": "",
			"operator": "",
			"in_use": False,
			"in_use_since": "",
			"delayed_logoff_in_progress": tool.tool_or_parent_id() in tools_with_delayed_logoff_in_effect,
			"problematic": False,
			"operational": tool.operational,
			"required_resource_is_unavailable": False,
			"nonrequired_resource_is_unavailable": False,
			"scheduled_outage": False,
			"scheduled_partial_outage": False,
			"area_name": tool.requires_area_access.name if tool.requires_area_access else None,
			"area_requires_reservation": tool.requires_area_access.requires_reservation
			if tool.requires_area_access
			else False,
		}
	for task in tasks:
		result[task.tool.id]["problematic"] = True
	for event in usage_events:
		result[event.tool.tool_or_parent_id()]["operator"] = str(event.operator)
		result[event.tool.tool_or_parent_id()]["user"] = str(event.operator)
		if event.user != event.operator:
			result[event.tool.tool_or_parent_id()]["user"] += " on behalf of " + str(event.user)
		result[event.tool.tool_or_parent_id()]["in_use"] = True
		result[event.tool.tool_or_parent_id()]["in_use_since"] = event.start
	for resource in unavailable_resources:
		for tool in resource.fully_dependent_tools.filter(visible=True):
			result[tool.id]["required_resource_is_unavailable"] = True
		for tool in resource.partially_dependent_tools.filter(visible=True):
			result[tool.id]["nonrequired_resource_is_unavailable"] = True
	for outage in scheduled_outages:
		if outage.tool_id and outage.tool.visible:
			result[outage.tool.id]["scheduled_outage"] = True
		elif outage.resource_id:
			for t in outage.resource.fully_dependent_tools.filter(visible=True):
				result[t.id]["scheduled_outage"] = True
			for t in outage.resource.partially_dependent_tools.filter(visible=True):
				result[t.id]["scheduled_partial_outage"] = True
	return result



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.utils import timezone
from django.views.decorators.http import require_GET

from NEMO.models import AreaAccessRecord, UsageEvent, Alert, Resource, Area
from NEMO.views.alerts import delete_expired_alerts


@login_required
@require_GET
def jumbotron(request):
	return render(request, 'jumbotron/jumbotron.html')


@login_required
@require_GET
def jumbotron_content(request):
	delete_expired_alerts()
	dictionary = {
		'facility_occupants': AreaAccessRecord.objects.filter(end=None, staff_charge=None).prefetch_related('customer', 'project').order_by('area__name', 'start'),
		'usage_events': UsageEvent.objects.filter(end=None).prefetch_related('operator', 'user', 'tool'),
		'alerts': Alert.objects.filter(user=None, debut_time__lte=timezone.now(), expired=False, deleted=False),
		'disabled_resources': Resource.objects.filter(available=False),
	}
	return render(request, 'jumbotron/jumbotron_content.html', dictionary)



from datetime import datetime, timedelta
from http import HTTPStatus

from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render
from django.urls import reverse
from django.utils import timezone
from django.utils.dateparse import parse_date, parse_time
from django.views.decorators.http import require_GET, require_POST

from NEMO.decorators import synchronized
from NEMO.exceptions import RequiredUnansweredQuestionsException
from NEMO.models import BadgeReader, Project, Reservation, ReservationItemType, Tool, UsageEvent, User
from NEMO.utilities import localize, quiet_int
from NEMO.views.calendar import (
	cancel_the_reservation,
	extract_configuration,
	extract_reservation_questions,
	render_reservation_questions,
	shorten_reservation,
)
from NEMO.views.policy import (
	check_policy_to_disable_tool,
	check_policy_to_enable_tool,
	check_policy_to_save_reservation,
)
from NEMO.views.status_dashboard import create_tool_summary
from NEMO.views.tool_control import (
	email_managers_required_questions_disable_tool,
	interlock_bypass_allowed,
	interlock_error,
)
from NEMO.widgets.dynamic_form import DynamicForm


@login_required
@permission_required("NEMO.kiosk")
@require_POST
def enable_tool(request):
	return do_enable_tool(request, request.POST["tool_id"])


@synchronized("tool_id")
def do_enable_tool(request, tool_id):
	tool = Tool.objects.get(id=tool_id)
	customer = User.objects.get(id=request.POST["customer_id"])
	project = Project.objects.get(id=request.POST["project_id"])
	bypass_interlock = request.POST.get("bypass", "False") == "True"

	response = check_policy_to_enable_tool(tool, operator=customer, user=customer, project=project, staff_charge=False)
	if response.status_code != HTTPStatus.OK:
		dictionary = {
			"message": "You are not authorized to enable this tool. {}".format(response.content.decode()),
			"delay": 10,
		}
		return render(request, "kiosk/acknowledgement.html", dictionary)

	# All policy checks passed so enable the tool for the user.
	if tool.interlock and not tool.interlock.unlock():
		if bypass_interlock and interlock_bypass_allowed(customer):
			pass
		else:
			return interlock_error("Enable", customer)

	# Create a new usage event to track how long the user uses the tool.
	new_usage_event = UsageEvent()
	new_usage_event.operator = customer
	new_usage_event.user = customer
	new_usage_event.project = project
	new_usage_event.tool = tool
	new_usage_event.save()

	dictionary = {"message": "You can now use the {}".format(tool), "badge_number": customer.badge_number}
	return render(request, "kiosk/acknowledgement.html", dictionary)


@login_required
@permission_required("NEMO.kiosk")
@require_POST
def disable_tool(request):
	return do_disable_tool(request, request.POST["tool_id"])


@synchronized("tool_id")
def do_disable_tool(request, tool_id):
	tool = Tool.objects.get(id=tool_id)
	customer = User.objects.get(id=request.POST["customer_id"])
	downtime = timedelta(minutes=quiet_int(request.POST.get("downtime")))
	bypass_interlock = request.POST.get("bypass", "False") == "True"
	response = check_policy_to_disable_tool(tool, customer, downtime)
	if response.status_code != HTTPStatus.OK:
		dictionary = {"message": response.content, "delay": 10}
		return render(request, "kiosk/acknowledgement.html", dictionary)

	# All policy checks passed so try to disable the tool for the user.
	if tool.interlock and not tool.interlock.lock():
		if bypass_interlock and interlock_bypass_allowed(customer):
			pass
		else:
			return interlock_error("Disable", customer)

	# Shorten the user's tool reservation since we are now done using the tool
	shorten_reservation(user=customer, item=tool, new_end=timezone.now() + downtime)

	# End the current usage event for the tool and save it.
	current_usage_event = tool.get_current_usage_event()
	current_usage_event.end = timezone.now() + downtime

	# Collect post-usage questions
	dynamic_form = DynamicForm(tool.post_usage_questions)

	try:
		current_usage_event.run_data = dynamic_form.extract(request)
	except RequiredUnansweredQuestionsException as e:
		if customer.is_staff and customer != current_usage_event.operator and current_usage_event.user != customer:
			# if a staff is forcing somebody off the tool and there are required questions, send an email and proceed
			current_usage_event.run_data = e.run_data
			email_managers_required_questions_disable_tool(current_usage_event.operator, customer, tool, e.questions)
		else:
			dictionary = {"message": str(e), "delay": 10}
			return render(request, "kiosk/acknowledgement.html", dictionary)

	dynamic_form.charge_for_consumables(
		current_usage_event.user,
		current_usage_event.operator,
		current_usage_event.project,
		current_usage_event.run_data,
		request,
	)
	dynamic_form.update_tool_counters(current_usage_event.run_data, tool.id)

	current_usage_event.save()
	dictionary = {"message": "You are no longer using the {}".format(tool), "badge_number": customer.badge_number}
	return render(request, "kiosk/acknowledgement.html", dictionary)


@login_required
@permission_required("NEMO.kiosk")
@require_POST
def reserve_tool(request):
	tool = Tool.objects.get(id=request.POST["tool_id"])
	customer = User.objects.get(id=request.POST["customer_id"])
	project = Project.objects.get(id=request.POST["project_id"])
	back = request.POST["back"]

	error_dictionary = {"back": back, "tool": tool, "project": project, "customer": customer}

	""" Create a reservation for a user. """
	try:
		date = parse_date(request.POST["date"])
		start = localize(datetime.combine(date, parse_time(request.POST["start"])))
		end = localize(datetime.combine(date, parse_time(request.POST["end"])))
	except:
		error_dictionary["message"] = "Please enter a valid date, start time, and end time for the reservation."
		return render(request, "kiosk/error.html", error_dictionary)
	# Create the new reservation:
	reservation = Reservation()
	reservation.project = project
	reservation.user = customer
	reservation.creator = customer
	reservation.reservation_item = tool
	reservation.start = start
	reservation.end = end
	reservation.short_notice = tool.determine_insufficient_notice(start)
	policy_problems, overridable = check_policy_to_save_reservation(
		cancelled_reservation=None,
		new_reservation=reservation,
		user_creating_reservation=customer,
		explicit_policy_override=False,
	)

	# If there was a problem in saving the reservation then return the error...
	if policy_problems:
		error_dictionary["message"] = policy_problems[0]
		return render(request, "kiosk/error.html", error_dictionary)

	# All policy checks have passed.
	if project is None and not customer.is_staff:
		error_dictionary["message"] = "You must specify a project for your reservation"
		return render(request, "kiosk/error.html", error_dictionary)

	reservation.additional_information, reservation.self_configuration = extract_configuration(request)
	# Reservation can't be short notice if the user is configuring the tool themselves.
	if reservation.self_configuration:
		reservation.short_notice = False

	# Reservation questions if applicable
	try:
		reservation.question_data = extract_reservation_questions(request, ReservationItemType.TOOL, tool.id, reservation.project)
	except RequiredUnansweredQuestionsException as e:
		error_dictionary["message"] = str(e)
		return render(request, "kiosk/error.html", error_dictionary)

	reservation.save_and_notify()
	return render(request, "kiosk/success.html", {"new_reservation": reservation, "customer": customer})


@login_required
@permission_required("NEMO.kiosk")
@require_POST
def cancel_reservation(request, reservation_id):
	""" Cancel a reservation for a user. """
	reservation = Reservation.objects.get(id=reservation_id)
	customer = User.objects.get(id=request.POST["customer_id"])

	response = cancel_the_reservation(reservation=reservation, user_cancelling_reservation=customer, reason=None)

	if response.status_code == HTTPStatus.OK:
		return render(request, "kiosk/success.html", {"cancelled_reservation": reservation, "customer": customer})
	else:
		return render(request, "kiosk/error.html", {"message": response.content, "customer": customer})


@login_required
@permission_required("NEMO.kiosk")
@require_POST
def tool_reservation(request, tool_id, user_id, back):
	tool = Tool.objects.get(id=tool_id, visible=True)
	customer = User.objects.get(id=user_id)
	project = Project.objects.get(id=request.POST["project_id"])

	dictionary = tool.get_configuration_information(user=customer, start=None)
	dictionary["tool"] = tool
	dictionary["date"] = None
	dictionary["project"] = project
	dictionary["customer"] = customer
	dictionary["back"] = back
	dictionary["tool_reservation_times"] = list(
		Reservation.objects.filter(cancelled=False, missed=False, shortened=False, tool=tool, start__gte=timezone.now())
	)

	# Reservation questions if applicable
	dictionary["reservation_questions"] = render_reservation_questions(ReservationItemType.TOOL, tool_id, project, True)

	return render(request, "kiosk/tool_reservation.html", dictionary)


@login_required
@permission_required("NEMO.kiosk")
@require_GET
def choices(request):
	try:
		customer = User.objects.get(badge_number=request.GET["badge_number"])
		usage_events = (
			UsageEvent.objects.filter(operator=customer.id, end=None)
				.order_by("tool__name")
				.prefetch_related("tool", "project")
		)
		tools_in_use = [u.tool.tool_or_parent_id() for u in usage_events]
		fifteen_minutes_from_now = timezone.now() + timedelta(minutes=15)
		tool_reservations = (
			Reservation.objects.filter(
				tool__isnull=False,
				end__gt=timezone.now(),
				user=customer,
				missed=False,
				cancelled=False,
				shortened=False,
			)
				.exclude(tool_id__in=tools_in_use, start__lte=fifteen_minutes_from_now)
				.exclude(ancestor__shortened=True)
				.order_by("start")
		)
	except:
		dictionary = {
			"message": "Your badge wasn't recognized. If you got a new one recently then we'll need to update your account. Please contact staff to resolve the problem."
		}
		return render(request, "kiosk/acknowledgement.html", dictionary)

	categories = [
		t[0] for t in Tool.objects.filter(visible=True).order_by("_category").values_list("_category").distinct()
	]
	unqualified_categories = [
		category
		for category in categories
		if not customer.is_staff
		   and not Tool.objects.filter(
			visible=True, _category=category, id__in=customer.qualifications.all().values_list("id")
		).exists()
	]
	dictionary = {
		"now": timezone.now(),
		"customer": customer,
		"usage_events": list(usage_events),
		"upcoming_reservations": tool_reservations,
		"tool_summary": create_tool_summary(),
		"categories": categories,
		"unqualified_categories": unqualified_categories,
	}
	return render(request, "kiosk/choices.html", dictionary)


@login_required
@permission_required("NEMO.kiosk")
@require_GET
def category_choices(request, category, user_id):
	try:
		customer = User.objects.get(id=user_id)
	except:
		dictionary = {
			"message": "Your badge wasn't recognized. If you got a new one recently then we'll need to update your account. Please contact staff to resolve the problem."
		}
		return render(request, "kiosk/acknowledgement.html", dictionary)
	tools = Tool.objects.filter(visible=True, _category=category)
	dictionary = {
		"customer": customer,
		"category": category,
		"tools": tools,
		"unqualified_tools": [
			tool for tool in tools if not customer.is_staff and tool not in customer.qualifications.all()
		],
		"tool_summary": create_tool_summary(),
	}
	return render(request, "kiosk/category_choices.html", dictionary)


@login_required
@permission_required("NEMO.kiosk")
@require_GET
def tool_information(request, tool_id, user_id, back):
	tool = Tool.objects.get(id=tool_id, visible=True)
	customer = User.objects.get(id=user_id)
	dictionary = {
		"customer": customer,
		"tool": tool,
		"rendered_configuration_html": tool.configuration_widget(customer),
		"post_usage_questions": DynamicForm(tool.post_usage_questions).render(
			"tool_usage_group_question", tool.id, virtual_inputs=True
		),
		"back": back,
	}
	try:
		current_reservation = Reservation.objects.get(
			start__lt=timezone.now(),
			end__gt=timezone.now(),
			cancelled=False,
			missed=False,
			shortened=False,
			user=customer,
			tool=tool,
		)
		remaining_reservation_duration = int((current_reservation.end - timezone.now()).total_seconds() / 60)
		# We don't need to bother telling the user their reservation will be shortened if there's less than two minutes left.
		# Staff are exempt from reservation shortening.
		if remaining_reservation_duration > 2 and not customer.is_staff:
			dictionary["remaining_reservation_duration"] = remaining_reservation_duration
	except Reservation.DoesNotExist:
		pass
	return render(request, "kiosk/tool_information.html", dictionary)


@login_required
@permission_required("NEMO.kiosk")
@require_GET
def kiosk(request, location=None):
	if location and Tool.objects.filter(_location=location, visible=True).exists():
		reader_id = request.GET.get("reader_id")
		dictionary = {
			"location": location,
			"badge_reader": BadgeReader.objects.get(id=reader_id) if reader_id else BadgeReader.default(),
		}
		return render(request, "kiosk/kiosk.html", dictionary)
	else:
		locations = sorted(list(set([tool.location for tool in Tool.objects.filter(visible=True)])))
		dictionary = {
			"locations": [
				{"url": reverse("kiosk", kwargs={"location": location}), "name": location} for location in locations
			]
		}
		return render(request, "kiosk/location_directory.html", dictionary)



# Generated by Django 2.2.10 on 2020-05-22 17:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0017_remove_facility_and_site_names_from_help_text'),
    ]

    operations = [
        migrations.AddField(
            model_name='area',
            name='count_staff_in_occupancy',
            field=models.BooleanField(default=True, help_text='Indicates that staff users will count towards maximum capacity.'),
        ),
    ]



import io
from datetime import datetime, timedelta
from http import HTTPStatus
from json import dumps, loads
from logging import getLogger
from re import match
from typing import List, Optional, Union

import pytz
from dateutil import rrule
from django.conf import settings
from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseForbidden, HttpResponseNotFound
from django.shortcuts import get_object_or_404, redirect, render
from django.template import Context, Template
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.views.decorators.http import require_GET, require_POST

from NEMO.decorators import disable_session_expiry_refresh, staff_member_required, synchronized
from NEMO.exceptions import ProjectChargeException, RequiredUnansweredQuestionsException
from NEMO.models import (
	Alert,
	Area,
	AreaAccessRecord,
	Closure,
	ClosureTime,
	Configuration,
	Project,
	Reservation,
	ReservationItemType,
	ReservationQuestions,
	ScheduledOutage,
	ScheduledOutageCategory,
	StaffCharge,
	Tool,
	UsageEvent,
	User,
)
from NEMO.utilities import (
	EmailCategory,
	as_timezone,
	bootstrap_primary_color,
	create_email_attachment,
	distinct_qs_value_list,
	extract_dates,
	extract_times,
	format_datetime,
	localize,
	parse_parameter_string,
	send_mail,
)
from NEMO.views.constants import ADDITIONAL_INFORMATION_MAXIMUM_LENGTH
from NEMO.views.customization import get_customization, get_media_file_contents
from NEMO.views.policy import (
	check_billing_to_project,
	check_policy_to_cancel_reservation,
	check_policy_to_create_outage,
	check_policy_to_save_reservation,
	check_tool_reservation_requiring_area,
	maximum_users_in_overlapping_reservations,
)
from NEMO.widgets.dynamic_form import DynamicForm, render_group_questions

calendar_logger = getLogger(__name__)

recurrence_frequency_display = {
	'DAILY': 'Day(s)',
	'DAILY_WEEKDAYS': 'Week Day(s)',
	'DAILY WEEKENDS': 'Weekend Day(s)',
	'WEEKLY': 'Week(s)',
	'MONTHLY': 'Month(s)',
}

recurrence_frequencies = {
	'DAILY': rrule.DAILY,
	'DAILY_WEEKDAYS': rrule.DAILY,
	'DAILY WEEKENDS': rrule.DAILY,
	'WEEKLY': rrule.WEEKLY,
	'MONTHLY': rrule.MONTHLY,
}


@login_required
@require_GET
def calendar(request, item_type=None, item_id=None):
	""" Present the calendar view to the user. """
	user:User = request.user
	if request.device == 'mobile':
		if item_type and item_type == 'tool' and item_id:
			return redirect('view_calendar', item_id)
		else:
			return redirect('choose_item', 'view_calendar')

	tools = Tool.objects.filter(visible=True).only('name', '_category', 'parent_tool_id').order_by('_category', 'name')
	areas = Area.objects.filter(requires_reservation=True).only('name')

	# We want to remove areas the user doesn't have access to
	display_all_areas = get_customization('calendar_display_not_qualified_areas') == 'enabled'
	if not display_all_areas and areas and user and not user.is_superuser:
		areas = [area for area in areas if area in user.accessible_areas()]

	from NEMO.widgets.item_tree import ItemTree
	rendered_item_tree_html = ItemTree().render(None, {'tools': tools, 'areas':areas, 'user': request.user})

	calendar_view = get_customization('calendar_view')
	calendar_first_day_of_week = get_customization('calendar_first_day_of_week')
	calendar_day_column_format = get_customization('calendar_day_column_format')
	calendar_week_column_format = get_customization('calendar_week_column_format')
	calendar_month_column_format = get_customization('calendar_month_column_format')
	calendar_start_of_the_day = get_customization('calendar_start_of_the_day')
	calendar_now_indicator = get_customization('calendar_now_indicator')
	calendar_all_tools = get_customization('calendar_all_tools')
	calendar_all_areas = get_customization('calendar_all_areas')
	calendar_all_areastools = get_customization('calendar_all_areastools')

	dictionary = {
		'rendered_item_tree_html': rendered_item_tree_html,
		'tools': list(tools),
		'areas': list(areas),
		'auto_select_item_id': item_id,
		'auto_select_item_type': item_type,
		'calendar_view' : calendar_view,
		'calendar_first_day_of_week' : calendar_first_day_of_week,
		'calendar_day_column_format' : calendar_day_column_format,
		'calendar_week_column_format' : calendar_week_column_format,
		'calendar_month_column_format' : calendar_month_column_format,
		'calendar_start_of_the_day' : calendar_start_of_the_day,
		'calendar_now_indicator' : calendar_now_indicator,
		'calendar_all_tools': calendar_all_tools,
		'calendar_all_areas': calendar_all_areas,
		'calendar_all_areastools': calendar_all_areastools,
		'self_login': False,
		'self_logout': False,
	}
	login_logout = get_customization('calendar_login_logout', raise_exception=False)
	self_login = get_customization('self_log_in', raise_exception=False)
	self_logout = get_customization('self_log_out', raise_exception=False)
	if login_logout == 'enabled':
		dictionary['self_login'] = self_login == 'enabled'
		dictionary['self_logout'] = self_logout == 'enabled'
	if request.user.is_staff:
		dictionary['users'] = User.objects.all()
	return render(request, 'calendar/calendar.html', dictionary)


@login_required
@require_GET
@disable_session_expiry_refresh
def event_feed(request):
	""" Get all reservations for a specific time-window. Optionally: filter by tool, area or user name. """
	try:
		start, end = extract_dates(request.GET)
	except Exception as e:
		return HttpResponseBadRequest('Invalid start or end time. ' + str(e))

	# We don't want to let someone hammer the database with phony calendar feed lookups.
	# Block any requests that have a duration of more than 8 weeks. The FullCalendar
	# should only ever request 6 weeks of data at a time (at most).
	if end - start > timedelta(weeks=8):
		return HttpResponseBadRequest("Calendar feed request has too long a duration: " + str(end - start))

	event_type = request.GET.get('event_type')

	facility_name = get_customization('facility_name')
	if event_type == 'reservations':
		return reservation_event_feed(request, start, end)
	elif event_type == f"{facility_name.lower()} usage":
		return usage_event_feed(request, start, end)
	# Only staff may request a specific user's history...
	elif event_type == 'specific user' and request.user.is_staff:
		user = get_object_or_404(User, id=request.GET.get('user'))
		return specific_user_feed(request, user, start, end)
	else:
		return HttpResponseBadRequest('Invalid event type or operation not authorized.')


def reservation_event_feed(request, start, end):
	events = Reservation.objects.filter(cancelled=False, missed=False, shortened=False)
	outages = ScheduledOutage.objects.none()
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	events = events.exclude(start__lt=start, end__lt=start)
	events = events.exclude(start__gt=end, end__gt=end)
	all_tools = request.GET.get('all_tools')
	all_areas = request.GET.get('all_areas')
	all_areastools = request.GET.get('all_areastools')

	# Filter events that only have to do with the relevant tool/area.
	item_type = request.GET.get('item_type')
	if all_tools:
		events = events.filter(area=None)
	elif all_areas:
		events = events.filter(tool=None)
	if item_type:
		item_type = ReservationItemType(item_type)
		item_id = request.GET.get('item_id')
		if item_id and not (all_tools or all_areas or all_areastools):
			events = events.filter(**{f'{item_type.value}__id': item_id})
			if item_type == ReservationItemType.TOOL:
				outages = ScheduledOutage.objects.filter(Q(tool=item_id) | Q(resource__fully_dependent_tools__in=[item_id]))
			elif item_type == ReservationItemType.AREA:
				outages = Area.objects.get(pk=item_id).scheduled_outage_queryset()

	# Exclude outages for which the following is true:
	# The outage starts and ends before the time-window, and...
	# The outage starts and ends after the time-window.
	outages = outages.exclude(start__lt=start, end__lt=start)
	outages = outages.exclude(start__gt=end, end__gt=end)

	# Filter events that only have to do with the current user.
	personal_schedule = request.GET.get('personal_schedule')
	if personal_schedule:
		events = events.filter(user=request.user)

	dictionary = {
		'events': events,
		'outages': outages,
		'personal_schedule': personal_schedule,
		'all_tools': all_tools,
		'all_areas': all_areas,
		'all_areastools': all_areastools,
	}
	return render(request, 'calendar/reservation_event_feed.html', dictionary)


def usage_event_feed(request, start, end):
	usage_events = UsageEvent.objects.none()
	area_access_events = AreaAccessRecord.objects.none()
	missed_reservations = Reservation.objects.none()

	item_id = request.GET.get('item_id')
	item_type = ReservationItemType(request.GET.get('item_type')) if request.GET.get('item_type') else None

	personal_schedule = request.GET.get('personal_schedule')
	all_areas = request.GET.get('all_areas')
	all_tools = request.GET.get('all_tools')
	all_areastools = request.GET.get('all_areastools')

	if personal_schedule:
		# Filter events that only have to do with the current user.
		# Display missed reservations, tool and area usage when 'personal schedule' is selected
		usage_events = UsageEvent.objects.filter(user=request.user)
		area_access_events = AreaAccessRecord.objects.filter(customer=request.user)
		missed_reservations = Reservation.objects.filter(missed=True, user=request.user)
	elif all_areas:
		area_access_events = AreaAccessRecord.objects.filter()
		missed_reservations = Reservation.objects.filter(missed=True, tool=None)
	elif all_tools:
		usage_events = UsageEvent.objects.filter()
		missed_reservations = Reservation.objects.filter(missed=True, area=None)
	elif all_areastools:
		usage_events = UsageEvent.objects.all()
		area_access_events = AreaAccessRecord.objects.filter()
		missed_reservations = Reservation.objects.filter(missed=True)
	elif item_type:
		reservation_filter = {item_type.value: item_id}
		missed_reservations = Reservation.objects.filter(missed=True).filter(**reservation_filter)
		# Filter events that only have to do with the relevant tool or area.
		if item_id and item_type == ReservationItemType.TOOL:
			usage_events = UsageEvent.objects.filter(tool__id__in=Tool.objects.get(pk=item_id).get_family_tool_ids())
		if item_id and item_type == ReservationItemType.AREA:
			area_access_events = AreaAccessRecord.objects.filter(area__id=item_id)

	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	usage_events = usage_events.exclude(start__lt=start, end__lt=start)
	usage_events = usage_events.exclude(start__gt=end, end__gt=end)
	area_access_events = area_access_events.exclude(start__lt=start, end__lt=start)
	area_access_events = area_access_events.exclude(start__gt=end, end__gt=end)
	missed_reservations = missed_reservations.exclude(start__lt=start, end__lt=start)
	missed_reservations = missed_reservations.exclude(start__gt=end, end__gt=end)

	dictionary = {
		'usage_events': usage_events,
		'area_access_events': area_access_events,
		'personal_schedule': personal_schedule,
		'missed_reservations': missed_reservations,
		'all_tools': all_tools,
		'all_areas': all_areas,
		'all_areastools': all_areastools,
	}
	return render(request, 'calendar/usage_event_feed.html', dictionary)


def specific_user_feed(request, user, start, end):
	# Find all tool usage events for a user.
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	usage_events = UsageEvent.objects.filter(user=user)
	usage_events = usage_events.exclude(start__lt=start, end__lt=start)
	usage_events = usage_events.exclude(start__gt=end, end__gt=end)

	# Find all area access events for a user.
	area_access_events = AreaAccessRecord.objects.filter(customer=user)
	area_access_events = area_access_events.exclude(start__lt=start, end__lt=start)
	area_access_events = area_access_events.exclude(start__gt=end, end__gt=end)

	# Find all reservations for the user that were not missed or cancelled.
	reservations = Reservation.objects.filter(user=user, missed=False, cancelled=False, shortened=False)
	reservations = reservations.exclude(start__lt=start, end__lt=start)
	reservations = reservations.exclude(start__gt=end, end__gt=end)

	# Find all missed reservations for the user.
	missed_reservations = Reservation.objects.filter(user=user, missed=True)
	missed_reservations = missed_reservations.exclude(start__lt=start, end__lt=start)
	missed_reservations = missed_reservations.exclude(start__gt=end, end__gt=end)

	dictionary = {
		'usage_events': usage_events,
		'area_access_events': area_access_events,
		'reservations': reservations,
		'missed_reservations': missed_reservations,
	}
	return render(request, 'calendar/specific_user_feed.html', dictionary)


@login_required
@require_POST
def create_reservation(request):
	""" Create a reservation for a user. """
	try:
		start, end = extract_times(request.POST)
		item_type = request.POST['item_type']
		item_id = request.POST.get('item_id')
	except Exception as e:
		return HttpResponseBadRequest(str(e))
	return create_item_reservation(request, request.user, start, end, ReservationItemType(item_type), item_id)


@synchronized("current_user")
def create_item_reservation(request, current_user, start, end, item_type: ReservationItemType, item_id):
	item = get_object_or_404(item_type.get_object_class(), id=item_id)
	explicit_policy_override = False
	if current_user.is_staff:
		try:
			user = User.objects.get(id=request.POST['impersonate'])
		except:
			user = current_user
		try:
			explicit_policy_override = request.POST['explicit_policy_override'] == 'true'
		except:
			pass
	else:
		user = current_user
	# Create the new reservation:
	new_reservation = Reservation()
	new_reservation.user = user
	new_reservation.creator = current_user
	# set tool or area
	setattr(new_reservation, item_type.value, item)
	new_reservation.start = start
	new_reservation.end = end
	new_reservation.short_notice = item.determine_insufficient_notice(start) if item_type == ReservationItemType.TOOL else False
	policy_problems, overridable = check_policy_to_save_reservation(cancelled_reservation=None, new_reservation=new_reservation, user_creating_reservation=request.user, explicit_policy_override=explicit_policy_override)

	# If there was a policy problem with the reservation then return the error...
	if policy_problems:
		return render(request, 'calendar/policy_dialog.html', {
			'policy_problems': policy_problems,
			'overridable': overridable and request.user.is_staff,
			'reservation_action': 'create'
		})

	# All policy checks have passed.

	# If the user only has one project then associate it with the reservation.
	# Otherwise, present a dialog box for the user to choose which project to associate.
	if not user.is_staff:
		active_projects = user.active_projects()
		if len(active_projects) == 1:
			new_reservation.project = active_projects[0]
		else:
			try:
				new_reservation.project = Project.objects.get(id=request.POST['project_id'])
			except:
				return render(request, 'calendar/project_choice.html', {
					'active_projects': active_projects,
					'missed_reservation_threshold': new_reservation.reservation_item.missed_reservation_threshold
				})

		# Check if we are allowed to bill to project
		try:
			check_billing_to_project(new_reservation.project, user, new_reservation.reservation_item)
		except ProjectChargeException as e:
			policy_problems.append(e.msg)
			return render(request, 'calendar/policy_dialog.html', {'policy_problems': policy_problems, 'overridable': False, 'reservation_action': 'create'})

	# Reservation questions if applicable
	reservation_questions = render_reservation_questions(item_type, item_id, new_reservation.project)
	if reservation_questions:
		if not bool(request.POST.get("reservation_questions", False)):
			# We have not yet asked the questions
			return render(request, 'calendar/reservation_questions.html', {'reservation_questions': reservation_questions})
		else:
			# We already asked before, now we need to extract the results
			try:
				new_reservation.question_data = extract_reservation_questions(request, item_type, item_id, new_reservation.project)
			except RequiredUnansweredQuestionsException as e:
				dictionary = {'error': str(e), 'reservation_questions': reservation_questions}
				return render(request, 'calendar/reservation_questions.html', dictionary)

	# Configuration rules only apply to tools
	if item_type == ReservationItemType.TOOL:
		configured = (request.POST.get('configured') == "true")
		# If a reservation is requested and the tool does not require configuration...
		if not item.is_configurable():
			new_reservation.save_and_notify()
			return reservation_success(request, new_reservation)

		# If a reservation is requested and the tool requires configuration that has not been submitted...
		elif item.is_configurable() and not configured:
			configuration_information = item.get_configuration_information(user=user, start=start)
			return render(request, 'calendar/configuration.html', configuration_information)

		# If a reservation is requested and configuration information is present also...
		elif item.is_configurable() and configured:
			new_reservation.additional_information, new_reservation.self_configuration = extract_configuration(request)
			# Reservation can't be short notice if the user is configuring the tool themselves.
			if new_reservation.self_configuration:
				new_reservation.short_notice = False
			new_reservation.save_and_notify()
			return reservation_success(request, new_reservation)

	elif item_type == ReservationItemType.AREA:
		new_reservation.save_and_notify()
		return HttpResponse()

	return HttpResponseBadRequest("Reservation creation failed because invalid parameters were sent to the server.")


def reservation_success(request, reservation: Reservation):
	""" Checks area capacity and display warning message if capacity is high """
	max_area_overlap, max_location_overlap = (0,0)
	max_area_time, max_location_time = (None, None)
	area: Area = reservation.tool.requires_area_access if reservation.reservation_item_type == ReservationItemType.TOOL else reservation.area
	location = reservation.tool.location if reservation.reservation_item_type == ReservationItemType.TOOL else None
	if area and area.reservation_warning:
		overlapping_reservations_in_same_area = Reservation.objects.filter(cancelled=False, missed=False, shortened=False, end__gte=reservation.start, start__lte=reservation.end)
		if reservation.reservation_item_type == ReservationItemType.TOOL:
			overlapping_reservations_in_same_area = overlapping_reservations_in_same_area.filter(tool__in=Tool.objects.filter(_requires_area_access=area))
		elif reservation.reservation_item_type == ReservationItemType.AREA:
			overlapping_reservations_in_same_area = overlapping_reservations_in_same_area.filter(area=area)
		max_area_overlap, max_area_time = maximum_users_in_overlapping_reservations(overlapping_reservations_in_same_area)
		if location:
			overlapping_reservations_in_same_location = overlapping_reservations_in_same_area.filter(tool__in=Tool.objects.filter(_location=location))
			max_location_overlap, max_location_time = maximum_users_in_overlapping_reservations(overlapping_reservations_in_same_location)
	if max_area_overlap and max_area_overlap >= area.warning_capacity():
		dictionary = {
			'area': area,
			'location': location,
			'max_area_count': max_area_overlap,
			'max_location_count': max_location_overlap,
			'max_area_time': max(max_area_time, reservation.start),
			'max_location_time': max(max_location_time, reservation.start) if max_location_time else None,
		}
		return render(request, 'calendar/reservation_warning.html', dictionary, status=201) # send 201 code CREATED to indicate success but with more information to come
	else:
		return HttpResponse()


def extract_configuration(request):
	cleaned_configuration = []
	for key, value in request.POST.items():
		entry = parse_configuration_entry(key, value)
		if entry:
			cleaned_configuration.append(entry)
	# Sort by configuration display priority and join the results:
	result = ''
	for config in sorted(cleaned_configuration):
		result += config[1] + '\n'
	if 'additional_information' in request.POST:
		result += request.POST['additional_information'][:ADDITIONAL_INFORMATION_MAXIMUM_LENGTH].strip()
	self_configuration = True if request.POST.get('self_configuration') == 'on' else False
	return result, self_configuration


def parse_configuration_entry(key, value):
	if value == "" or not match("^configuration_[0-9]+__slot_[0-9]+__display_priority_[0-9]+$", key):
		return None
	config_id, slot, display_priority = [int(s) for s in key.split('_') if s.isdigit()]
	configuration = Configuration.objects.get(pk=config_id)
	available_setting = configuration.get_available_setting(value)
	if len(configuration.current_settings_as_list()) == 1:
		return display_priority, configuration.name + " needs to be set to " + available_setting + "."
	else:
		return display_priority, configuration.configurable_item_name + " #" + str(slot + 1) + " needs to be set to " + available_setting + "."


@staff_member_required
@require_POST
def create_outage(request):
	""" Create an outage. """
	try:
		start, end = extract_times(request.POST)
		item_type = ReservationItemType(request.POST['item_type'])
		item_id = request.POST.get('item_id')
	except Exception as e:
		return HttpResponseBadRequest(str(e))
	item = get_object_or_404(item_type.get_object_class(), id=item_id)
	# Create the new reservation:
	outage = ScheduledOutage()
	outage.creator = request.user
	outage.category = request.POST.get('category', '')[:200]
	outage.outage_item = item
	outage.start = start
	outage.end = end

	# If there is a policy problem for the outage then return the error...
	policy_problem = check_policy_to_create_outage(outage)
	if policy_problem:
		return HttpResponseBadRequest(policy_problem)

	# Make sure there is at least an outage title
	if not request.POST.get('title'):
		calendar_outage_recurrence_limit = get_customization("calendar_outage_recurrence_limit")
		dictionary = {
			'categories': ScheduledOutageCategory.objects.all(),
			'recurrence_intervals': recurrence_frequency_display,
			'recurrence_date_start': start.date(),
			'calendar_outage_recurrence_limit': calendar_outage_recurrence_limit,
		}
		return render(request, 'calendar/scheduled_outage_information.html', dictionary)

	outage.title = request.POST['title']
	outage.details = request.POST.get('details', '')

	if request.POST.get('recurring_outage') == 'on':
		# we have to remove tz before creating rules otherwise 8am would become 7am after DST change for example.
		start_no_tz = outage.start.replace(tzinfo=None)
		end_no_tz = outage.end.replace(tzinfo=None)

		submitted_frequency = request.POST.get('recurrence_frequency')
		submitted_date_until = request.POST.get('recurrence_until', None)
		date_until = end.replace(hour=0, minute=0, second=0)
		if submitted_date_until:
			date_until = localize(datetime.strptime(submitted_date_until, '%m/%d/%Y'))
		date_until += timedelta(days=1, seconds=-1) # set at the end of the day
		by_week_day = None
		if submitted_frequency == 'DAILY_WEEKDAYS':
			by_week_day = (rrule.MO, rrule.TU, rrule.WE, rrule.TH, rrule.FR)
		elif submitted_frequency == 'DAILY_WEEKENDS':
			by_week_day = (rrule.SA, rrule.SU)
		frequency = recurrence_frequencies.get(submitted_frequency, rrule.DAILY)
		rules = rrule.rrule(dtstart=start, freq=frequency, interval=int(request.POST.get('recurrence_interval',1)), until=date_until, byweekday=by_week_day)
		for rule in list(rules):
			recurring_outage = ScheduledOutage()
			recurring_outage.creator = outage.creator
			recurring_outage.category = outage.category
			recurring_outage.outage_item = outage.outage_item
			recurring_outage.title = outage.title
			recurring_outage.details = outage.details
			recurring_outage.start = localize(start_no_tz.replace(year=rule.year, month=rule.month, day=rule.day))
			recurring_outage.end = localize(end_no_tz.replace(year=rule.year, month=rule.month, day=rule.day))
			recurring_outage.save()
	else:
		outage.save()

	return HttpResponse()


@login_required
@require_POST
def resize_reservation(request):
	""" Resize a reservation for a user. """
	try:
		delta = timedelta(minutes=int(request.POST['delta']))
	except:
		return HttpResponseBadRequest('Invalid delta')
	return modify_reservation(request, None, delta)


@staff_member_required
@require_POST
def resize_outage(request):
	""" Resize an outage """
	try:
		delta = timedelta(minutes=int(request.POST['delta']))
	except:
		return HttpResponseBadRequest('Invalid delta')
	return modify_outage(request, None, delta)


@login_required
@require_POST
def move_reservation(request):
	""" Move a reservation for a user. """
	try:
		delta = timedelta(minutes=int(request.POST['delta']))
	except:
		return HttpResponseBadRequest('Invalid delta')
	return modify_reservation(request, delta, delta)


@staff_member_required
@require_POST
def move_outage(request):
	""" Move a reservation for a user. """
	try:
		delta = timedelta(minutes=int(request.POST['delta']))
	except:
		return HttpResponseBadRequest('Invalid delta')
	return modify_outage(request, delta, delta)


def modify_reservation(request, start_delta, end_delta):
	"""
	Cancel the user's old reservation and create a new one. Reservations are cancelled and recreated so that
	reservation abuse can be tracked if necessary. This function should be called by other views and should
	not be tied directly to a URL.
	"""
	try:
		reservation_to_cancel = Reservation.objects.get(pk=request.POST.get('id'))
	except Reservation.DoesNotExist:
		return HttpResponseNotFound("The reservation that you wish to modify doesn't exist!")
	explicit_policy_override = False
	try:
		explicit_policy_override = request.POST['explicit_policy_override'] == 'true'
	except:
		pass
	# Record the current time so that the timestamp of the cancelled reservation and the new reservation match exactly.
	now = timezone.now()
	# Create a new reservation for the user by copying the original one.
	new_start = reservation_to_cancel.start + start_delta if start_delta else None
	new_end = reservation_to_cancel.end + end_delta if end_delta else None
	new_reservation = reservation_to_cancel.copy(new_start, new_end)
	# Set new creator/time
	new_reservation.creation_time = now
	new_reservation.creator = request.user

	response = check_policy_to_cancel_reservation(request.user, reservation_to_cancel, new_reservation)
	# Do not move the reservation if the user was not authorized to cancel it.
	if response.status_code != HTTPStatus.OK:
		return response

	# Cancel the user's original reservation.
	reservation_to_cancel.cancelled = True
	reservation_to_cancel.cancellation_time = now
	reservation_to_cancel.cancelled_by = request.user

	policy_problems, overridable = check_policy_to_save_reservation(cancelled_reservation=reservation_to_cancel, new_reservation=new_reservation, user_creating_reservation=request.user, explicit_policy_override=explicit_policy_override)
	if policy_problems:
		reservation_action = "resize" if start_delta is None else "move"
		return render(request, 'calendar/policy_dialog.html', {'policy_problems': policy_problems, 'overridable': overridable and request.user.is_staff, 'reservation_action': reservation_action})
	else:
		# All policy checks passed, so save the reservation.
		new_reservation.save_and_notify()
		reservation_to_cancel.descendant = new_reservation
		reservation_to_cancel.save_and_notify()
	return reservation_success(request, new_reservation)


def modify_outage(request, start_delta, end_delta):
	try:
		outage = ScheduledOutage.objects.get(pk=request.POST.get('id'))
	except ScheduledOutage.DoesNotExist:
		return HttpResponseNotFound("The outage that you wish to modify doesn't exist!")
	if start_delta:
		outage.start += start_delta
	outage.end += end_delta
	policy_problem = check_policy_to_create_outage(outage)
	if policy_problem:
		return HttpResponseBadRequest(policy_problem)
	else:
		# All policy checks passed, so save the reservation.
		outage.save()
	return HttpResponse()


@login_required
@require_POST
def cancel_reservation(request, reservation_id):
	""" Cancel a reservation for a user. """
	reservation = get_object_or_404(Reservation, id=reservation_id)

	reason = parse_parameter_string(request.POST, 'reason')
	response = cancel_the_reservation(reservation=reservation, user_cancelling_reservation=request.user, reason=reason)

	if request.device == 'desktop':
		return response
	if request.device == 'mobile':
		if response.status_code == HTTPStatus.OK:
			return render(request, 'mobile/cancellation_result.html', {'event_type': 'Reservation', 'tool': reservation.tool})
		else:
			return render(request, 'mobile/error.html', {'message': response.content})


@staff_member_required
@require_POST
def cancel_outage(request, outage_id):
	outage = get_object_or_404(ScheduledOutage, id=outage_id)
	outage.delete()
	if request.device == 'desktop':
		return HttpResponse()
	if request.device == 'mobile':
		dictionary = {'event_type': 'Scheduled outage', 'tool': outage.tool, 'area': outage.area}
		return render(request, 'mobile/cancellation_result.html', dictionary)


@staff_member_required
@require_POST
def set_reservation_title(request, reservation_id):
	""" Change reservation title for a user. """
	reservation = get_object_or_404(Reservation, id=reservation_id)
	reservation.title = request.POST.get('title', '')[:reservation._meta.get_field('title').max_length]
	reservation.save()
	return HttpResponse()


@login_required
@require_POST
def change_reservation_project(request, reservation_id):
	""" Change reservation project for a user. """
	reservation = get_object_or_404(Reservation, id=reservation_id)
	project = get_object_or_404(Project, id=request.POST['project_id'])
	try:
		check_billing_to_project(project, reservation.user, reservation.reservation_item)
	except ProjectChargeException as e:
		return HttpResponseBadRequest(e.msg)

	if (request.user.is_staff or request.user == reservation.user) and \
		reservation.has_not_ended() and reservation.has_not_started() and \
		project in reservation.user.active_projects():
		reservation.project = project
		reservation.save()
	else:
		# project for reservation was not eligible to be changed
		if not (request.user.is_staff or request.user == reservation.user):
			return HttpResponseForbidden(f"{request.user} is not authorized to change the project for this reservation")
		if not reservation.has_not_ended():
			return HttpResponseBadRequest("Project cannot be changed; reservation has already ended")
		if not reservation.has_not_started():
			return HttpResponseBadRequest("Project cannot be changed; reservation has already started")
		if project not in reservation.user.active_projects():
			return HttpResponseForbidden(f"{project} is not one of {reservation.user}'s active projects")
	return HttpResponse()


@login_required
@permission_required('NEMO.trigger_timed_services', raise_exception=True)
@require_GET
def email_reservation_reminders(request):
	return send_email_reservation_reminders()


def send_email_reservation_reminders():
	# Exit early if the reservation reminder email template has not been customized for the organization yet.
	reservation_reminder_message = get_media_file_contents('reservation_reminder_email.html')
	reservation_warning_message = get_media_file_contents('reservation_warning_email.html')
	if not reservation_reminder_message or not reservation_warning_message:
		calendar_logger.error("Reservation reminder email couldn't be send because either reservation_reminder_email.html or reservation_warning_email.html is not defined")
		return HttpResponseNotFound('The reservation reminder and/or warning email templates have not been customized for your organization yet. Please visit the customization page to upload both templates, then reservation reminder email notifications can be sent.')

	# Find all reservations that are two hours from now, plus or minus 5 minutes to allow for time skew.
	preparation_time = 120
	tolerance = 5
	earliest_start = timezone.now() + timedelta(minutes=preparation_time) - timedelta(minutes=tolerance)
	latest_start = timezone.now() + timedelta(minutes=preparation_time) + timedelta(minutes=tolerance)
	upcoming_reservations = Reservation.objects.filter(cancelled=False, start__gt=earliest_start, start__lt=latest_start)
	# Email a reminder to each user with an upcoming reservation.
	for reservation in upcoming_reservations:
		item = reservation.reservation_item
		item_type = reservation.reservation_item_type
		if item_type == ReservationItemType.TOOL and item.operational and not item.problematic() and item.all_resources_available()\
				or item_type == ReservationItemType.AREA and not item.required_resource_is_unavailable():
			subject = item.name + " reservation reminder"
			rendered_message = Template(reservation_reminder_message).render(Context({'reservation': reservation, 'template_color': bootstrap_primary_color('success')}))
		elif (item_type == ReservationItemType.TOOL and not item.operational) or item.required_resource_is_unavailable():
			subject = item.name + " reservation problem"
			rendered_message = Template(reservation_warning_message).render(Context({'reservation': reservation, 'template_color': bootstrap_primary_color('danger'), 'fatal_error': True}))
		else:
			subject = item.name + " reservation warning"
			rendered_message = Template(reservation_warning_message).render(Context({'reservation': reservation, 'template_color': bootstrap_primary_color('warning'), 'fatal_error': False}))
		user_office_email = get_customization('user_office_email_address')
		reservation.user.email_user(subject=subject, content=rendered_message, from_email=user_office_email, email_category=EmailCategory.TIMED_SERVICES)
	return HttpResponse()


@login_required
@permission_required('NEMO.trigger_timed_services', raise_exception=True)
@require_GET
def email_reservation_ending_reminders(request):
	return send_email_reservation_ending_reminders()


def send_email_reservation_ending_reminders():
	# Exit early if the reservation ending reminder email template has not been customized for the organization yet.
	reservation_ending_reminder_message = get_media_file_contents('reservation_ending_reminder_email.html')
	user_office_email = get_customization('user_office_email_address')
	if not reservation_ending_reminder_message:
		calendar_logger.error("Reservation ending reminder email couldn't be send because either reservation_ending_reminder_email.html is not defined")
		return HttpResponseNotFound('The reservation ending reminder template has not been customized for your organization yet. Please visit the customization page to upload one, then reservation ending reminder email notifications can be sent.')

	# We only send ending reservation reminders to users that are currently logged in
	current_logged_in_user = AreaAccessRecord.objects.filter(end=None, staff_charge=None, customer__is_staff=False)
	valid_reservations = Reservation.objects.filter(cancelled=False, missed=False, shortened=False)
	user_area_reservations = valid_reservations.filter(area__isnull=False, user__in=current_logged_in_user.values_list('customer', flat=True))

	# Find all reservations that end 30 or 15 min from now, plus or minus 3 minutes to allow for time skew.
	reminder_times = [30,15]
	tolerance = 3
	time_filter = Q()
	for reminder_time in reminder_times:
		earliest_end = timezone.now() + timedelta(minutes=reminder_time) - timedelta(minutes=tolerance)
		latest_end = timezone.now() + timedelta(minutes=reminder_time) + timedelta(minutes=tolerance)
		new_filter = Q(end__gt=earliest_end, end__lt=latest_end)
		time_filter = time_filter | new_filter
	ending_reservations = user_area_reservations.filter(time_filter)
	# Email a reminder to each user with an reservation ending soon.
	for reservation in ending_reservations:
		subject = reservation.reservation_item.name + " reservation ending soon"
		rendered_message = Template(reservation_ending_reminder_message).render(Context({'reservation': reservation}))
		reservation.user.email_user(subject=subject, content=rendered_message, from_email=user_office_email, email_category=EmailCategory.TIMED_SERVICES)
	return HttpResponse()


@login_required
@permission_required('NEMO.trigger_timed_services', raise_exception=True)
@require_GET
def email_usage_reminders(request):
	projects_to_exclude = request.GET.getlist("projects_to_exclude[]")
	return send_email_usage_reminders(projects_to_exclude)


def send_email_usage_reminders(projects_to_exclude=None):
	if projects_to_exclude is None:
		projects_to_exclude = []
	busy_users = AreaAccessRecord.objects.filter(end=None, staff_charge=None).exclude(project__id__in=projects_to_exclude)
	busy_tools = UsageEvent.objects.filter(end=None).exclude(project__id__in=projects_to_exclude)

	# Make lists of all the things a user is logged in to.
	# We don't want to send 3 separate emails if a user is logged into three things.
	# Just send one email for all the things!
	aggregate = {}
	for access_record in busy_users:
		key = str(access_record.customer)
		aggregate[key] = {
			'email': access_record.customer.email,
			'first_name': access_record.customer.first_name,
			'resources_in_use': [access_record.area.name],
		}
	for usage_event in busy_tools:
		key = str(usage_event.operator)
		if key in aggregate:
			aggregate[key]['resources_in_use'].append(usage_event.tool.name)
		else:
			aggregate[key] = {
				'email': usage_event.operator.email,
				'first_name': usage_event.operator.first_name,
				'resources_in_use': [usage_event.tool.name],
			}

	user_office_email = get_customization('user_office_email_address')

	message = get_media_file_contents('usage_reminder_email.html')
	facility_name = get_customization('facility_name')
	if message:
		subject = f"{facility_name} usage"
		for user in aggregate.values():
			rendered_message = Template(message).render(Context({'user': user}))
			send_mail(subject=subject, content=rendered_message, from_email=user_office_email, to=[user['email']], email_category=EmailCategory.TIMED_SERVICES)

	message = get_media_file_contents('staff_charge_reminder_email.html')
	if message:
		busy_staff = StaffCharge.objects.filter(end=None)
		for staff_charge in busy_staff:
			subject = "Active staff charge since " + format_datetime(staff_charge.start)
			rendered_message = Template(message).render(Context({'staff_charge': staff_charge}))
			staff_charge.staff_member.email_user(subject=subject, content=rendered_message, from_email=user_office_email, email_category=EmailCategory.TIMED_SERVICES)

	return HttpResponse()


@login_required
@require_GET
@permission_required('NEMO.trigger_timed_services', raise_exception=True)
def cancel_unused_reservations(request):
	return do_cancel_unused_reservations()


def do_cancel_unused_reservations():
	"""
	Missed reservation for tools is when there is no tool activity during the reservation time + missed reservation threshold.
	Any tool usage will count, since we don't want to charge for missed reservation when users swap reservation or somebody else gets to use the tool.

	Missed reservation for areas is when there is no area access login during the reservation time + missed reservation threshold
	"""

	# Missed Tool Reservations
	tools = Tool.objects.filter(visible=True, _operational=True, _missed_reservation_threshold__isnull=False)
	missed_reservations = []
	for tool in tools:
		# If a tool is in use then there's no need to look for unused reservation time.
		if tool.in_use() or tool.required_resource_is_unavailable() or tool.scheduled_outage_in_progress():
			continue
		# Calculate the timestamp of how long a user can be late for a reservation.
		threshold = (timezone.now() - timedelta(minutes=tool.missed_reservation_threshold))
		threshold = datetime.replace(threshold, second=0, microsecond=0)  # Round down to the nearest minute.
		# Find the reservations that began exactly at the threshold.
		reservation = Reservation.objects.filter(cancelled=False, missed=False, shortened=False, tool=tool, user__is_staff=False, start=threshold, end__gt=timezone.now())
		for r in reservation:
			# Staff may abandon reservations.
			if r.user.is_staff:
				continue
			# If there was no tool enable or disable event since the threshold timestamp then we assume the reservation has been missed.
			if not (UsageEvent.objects.filter(tool_id__in=tool.get_family_tool_ids(), start__gte=threshold).exists() or UsageEvent.objects.filter(tool_id__in=tool.get_family_tool_ids(), end__gte=threshold).exists()):
				# Mark the reservation as missed and notify the user & staff.
				r.missed = True
				r.save()
				missed_reservations.append(r)

	# Missed Area Reservations
	areas = Area.objects.filter(missed_reservation_threshold__isnull=False)
	for area in areas:
		# if area has outage or required resource is unavailable, no need to look
		if area.required_resource_is_unavailable() or area.scheduled_outage_in_progress():
			continue

		# Calculate the timestamp of how long a user can be late for a reservation.
		threshold = (timezone.now() - timedelta(minutes=area.missed_reservation_threshold))
		threshold = datetime.replace(threshold, second=0, microsecond=0)  # Round down to the nearest minute.
		# Find the reservations that began exactly at the threshold.
		reservation = Reservation.objects.filter(cancelled=False, missed=False, shortened=False, area=area, user__is_staff=False, start=threshold, end__gt=timezone.now())
		for r in reservation:
			# Staff may abandon reservations.
			if r.user.is_staff:
				continue
			# if the user is not already logged in or if there was no area access starting or ending since the threshold timestamp then we assume the reservation was missed
			if not (AreaAccessRecord.objects.filter(area__id=area.id, customer=r.user, staff_charge=None, end=None).exists() or AreaAccessRecord.objects.filter(area__id=area.id, customer=r.user, start__gte=threshold).exists() or AreaAccessRecord.objects.filter(area__id=area.id, customer=r.user, end__gte=threshold).exists()):
				# Mark the reservation as missed and notify the user & staff.
				r.missed = True
				r.save()
				missed_reservations.append(r)

	for r in missed_reservations:
		send_missed_reservation_notification(r)

	return HttpResponse()


@login_required
@require_GET
@permission_required('NEMO.trigger_timed_services', raise_exception=True)
def email_out_of_time_reservation_notification(request):
	return send_email_out_of_time_reservation_notification()


def send_email_out_of_time_reservation_notification():
	"""
	Out of time reservation notification for areas is when a user is still logged in a area but his reservation expired.
	"""
	# Exit early if the out of time reservation email template has not been customized for the organization yet.
	# This feature only sends emails, so there if the template is not defined there nothing to do.
	if not get_media_file_contents('out_of_time_reservation_email.html'):
		return HttpResponseNotFound('The out of time reservation email template has not been customized for your organization yet. Please visit the customization page to upload a template, then out of time email notifications can be sent.')

	out_of_time_user_area = []

	# Find all logged users
	access_records:List[AreaAccessRecord] = AreaAccessRecord.objects.filter(end=None, staff_charge=None).prefetch_related('customer', 'area').only('customer', 'area')
	for access_record in access_records:
		# staff and service personnel are exempt from out of time notification
		customer = access_record.customer
		area = access_record.area
		if customer.is_staff or customer.is_service_personnel:
			continue

		if area.requires_reservation:
			# Calculate the timestamp of how late a user can be logged in after a reservation ended.
			threshold = timezone.now() if not area.logout_grace_period else timezone.now() - timedelta(minutes=area.logout_grace_period)
			threshold = datetime.replace(threshold, second=0, microsecond=0)  # Round down to the nearest minute.
			ending_reservations = Reservation.objects.filter(cancelled=False, missed=False, shortened=False, area=area, user=customer, start__lte=timezone.now(), end=threshold)
			# find out if a reservation is starting right at the same time (in case of back to back reservations, in which case customer is good)
			starting_reservations = Reservation.objects.filter(cancelled=False, missed=False, shortened=False, area=area, user=customer, start=threshold)
			if ending_reservations.exists() and not starting_reservations.exists():
				out_of_time_user_area.append(ending_reservations[0])

	for reservation in out_of_time_user_area:
		send_out_of_time_reservation_notification(reservation)

	return HttpResponse()


@staff_member_required
@require_GET
def proxy_reservation(request):
	return render(request, 'calendar/proxy_reservation.html', {'users': User.objects.filter(is_active=True)})


@login_required
@require_GET
def reservation_group_question(request, reservation_question_id, group_name):
	reservation_questions = get_object_or_404(ReservationQuestions, id=reservation_question_id)
	return HttpResponse(render_group_questions(request, reservation_questions.questions, "reservation_group_question", reservation_question_id, group_name))


def get_and_combine_reservation_questions(item_type: ReservationItemType, item_id: int, project: Project = None) -> List[ReservationQuestions]:
	reservation_questions = ReservationQuestions.objects.all()
	if item_type == ReservationItemType.TOOL:
		reservation_questions = reservation_questions.filter(tool_reservations=True)
		reservation_questions = reservation_questions.filter(Q(only_for_tools=None) | Q(only_for_tools__in=[item_id]))
	if item_type == ReservationItemType.AREA:
		reservation_questions = reservation_questions.filter(area_reservations=True)
		reservation_questions = reservation_questions.filter(Q(only_for_areas=None) | Q(only_for_areas__in=[item_id]))
	if project:
		reservation_questions = reservation_questions.filter(Q(only_for_projects=None) | Q(only_for_projects__in=[project.id]))
	else:
		reservation_questions = reservation_questions.filter(only_for_projects=None)
	return reservation_questions


def render_reservation_questions(item_type: ReservationItemType, item_id: int, project: Project = None, virtual_inputs: bool = False) -> str:
	reservation_questions = get_and_combine_reservation_questions(item_type, item_id, project)
	rendered_questions = ""
	for reservation_question in reservation_questions:
		rendered_questions += DynamicForm(reservation_question.questions).render("reservation_group_question", reservation_question.id, virtual_inputs)
	return mark_safe(rendered_questions)


def extract_reservation_questions(request, item_type: ReservationItemType, item_id: int, project: Project = None) -> str:
	reservation_questions = get_and_combine_reservation_questions(item_type, item_id, project)
	reservation_questions_json = []
	for reservation_question in reservation_questions:
		reservation_questions_json.extend(loads(reservation_question.questions))
	return DynamicForm(dumps(reservation_questions_json)).extract(request) if len(reservation_questions_json) else ""


def shorten_reservation(user: User, item: Union[Area, Tool], new_end: datetime = None):
	try:
		if new_end is None:
			new_end = timezone.now()
		current_reservation_qs = Reservation.objects.filter(start__lt=timezone.now(), end__gt=timezone.now(),
															cancelled=False, missed=False, shortened=False, user=user)
		current_reservation = current_reservation_qs.get(**{ReservationItemType.from_item(item).value: item})
		# Staff are exempt from mandatory reservation shortening.
		if user.is_staff is False:
			new_reservation = current_reservation.copy(new_end=new_end)
			new_reservation.save()
			current_reservation.shortened = True
			current_reservation.descendant = new_reservation
			current_reservation.save()
	except Reservation.DoesNotExist:
		pass


def cancel_the_reservation(reservation: Reservation, user_cancelling_reservation: User, reason: Optional[str]):
	# Check policy to cancel reservation contains rules common to cancelling and modifying
	response = check_policy_to_cancel_reservation(user_cancelling_reservation, reservation)

	# The following rules apply only for proper cancellation, not for modification
	# Staff must provide a reason when cancelling a reservation they do not own.
	if reservation.user != user_cancelling_reservation and not reason:
		response = HttpResponseBadRequest("You must provide a reason when cancelling someone else's reservation.")

	policy_problems = []
	check_tool_reservation_requiring_area(policy_problems, user_cancelling_reservation, reservation, None)
	if policy_problems:
		return HttpResponseBadRequest(policy_problems[0])

	if response.status_code == HTTPStatus.OK:
		# All policy checks passed, so cancel the reservation.
		reservation.cancelled = True
		reservation.cancellation_time = timezone.now()
		reservation.cancelled_by = user_cancelling_reservation

		if reason:
			''' don't notify (just save) in this case since we are sending a specific email for the cancellation '''
			reservation.save()
			email_contents = get_media_file_contents('cancellation_email.html')
			if email_contents:
				dictionary = {
					'staff_member': user_cancelling_reservation,
					'reservation': reservation,
					'reason': reason,
					'template_color': bootstrap_primary_color('info')
				}
				cancellation_email = Template(email_contents).render(Context(dictionary))
				recipients = [reservation.user.email]
				if reservation.area:
					recipients.extend(reservation.area.reservation_email_list())
				if reservation.user.get_preferences().attach_cancelled_reservation:
					attachment = create_ics_for_reservation(reservation, cancelled=True)
					send_mail(subject='Your reservation was cancelled', content=cancellation_email, from_email=user_cancelling_reservation.email, to=recipients, attachments=[attachment])
				else:
					send_mail(subject='Your reservation was cancelled', content=cancellation_email, from_email=user_cancelling_reservation.email, to=recipients)

		else:
			''' here the user cancelled his own reservation so notify him '''
			reservation.save_and_notify()

	return response


def send_missed_reservation_notification(reservation):
	message = get_media_file_contents('missed_reservation_email.html')
	user_office_email = get_customization('user_office_email_address')
	abuse_email = get_customization('abuse_email_address')
	if message and user_office_email:
		subject = "Missed reservation for the " + str(reservation.reservation_item)
		message = Template(message).render(Context({'reservation': reservation}))
		send_mail(subject=subject, content=message, from_email=user_office_email, to=[reservation.user.email, abuse_email, user_office_email], email_category=EmailCategory.TIMED_SERVICES)
	else:
		calendar_logger.error("Missed reservation email couldn't be send because missed_reservation_email.html or user_office_email are not defined")


def send_out_of_time_reservation_notification(reservation:Reservation):
	message = get_media_file_contents('out_of_time_reservation_email.html')
	user_office_email = get_customization('user_office_email_address')
	if message and user_office_email:
		subject = "Out of time in the " + str(reservation.area.name)
		message = Template(message).render(Context({'reservation': reservation}))
		recipients = [reservation.user.email]
		recipients.extend(reservation.area.abuse_email_list())
		send_mail(subject=subject, content=message, from_email=user_office_email, to=recipients, email_category=EmailCategory.TIMED_SERVICES)
	else:
		calendar_logger.error("Out of time reservation email couldn't be send because out_of_time_reservation_email.html or user_office_email are not defined")


def send_user_created_reservation_notification(reservation: Reservation):
	site_title = get_customization('site_title')
	recipients = [reservation.user.email] if reservation.user.get_preferences().attach_created_reservation else []
	if reservation.area:
		recipients.extend(reservation.area.reservation_email_list())
	if recipients:
		subject = f"[{site_title}] Reservation for the " + str(reservation.reservation_item)
		message = get_media_file_contents('reservation_created_user_email.html')
		message = Template(message).render(Context({'reservation': reservation}))
		user_office_email = get_customization('user_office_email_address')
		# We don't need to check for existence of reservation_created_user_email because we are attaching the ics reservation and sending the email regardless (message will be blank)
		if user_office_email:
			attachment = create_ics_for_reservation(reservation)
			send_mail(subject=subject, content=message, from_email=user_office_email, to=recipients, attachments=[attachment])
		else:
			calendar_logger.error("User created reservation notification could not be send because user_office_email_address is not defined")


def send_user_cancelled_reservation_notification(reservation: Reservation):
	site_title = get_customization('site_title')
	recipients = [reservation.user.email] if reservation.user.get_preferences().attach_cancelled_reservation else []
	if reservation.area:
		recipients.extend(reservation.area.reservation_email_list())
	if recipients:
		subject = f"[{site_title}] Cancelled Reservation for the " + str(reservation.reservation_item)
		message = get_media_file_contents('reservation_cancelled_user_email.html')
		message = Template(message).render(Context({'reservation': reservation}))
		user_office_email = get_customization('user_office_email_address')
		# We don't need to check for existence of reservation_cancelled_user_email because we are attaching the ics reservation and sending the email regardless (message will be blank)
		if user_office_email:
			attachment = create_ics_for_reservation(reservation, cancelled=True)
			send_mail(subject=subject, content=message, from_email=user_office_email, to=recipients, attachments=[attachment])
		else:
			calendar_logger.error("User cancelled reservation notification could not be send because user_office_email_address is not defined")


def create_ics_for_reservation(reservation: Reservation, cancelled=False):
	site_title = get_customization('site_title')
	reservation_organizer_email = getattr(settings, "RESERVATION_ORGANIZER_EMAIL", "no_reply")
	reservation_organizer = getattr(settings, "RESERVATION_ORGANIZER", site_title)
	method_name = 'CANCEL' if cancelled else 'REQUEST'
	method = f'METHOD:{method_name}\n'
	status = 'STATUS:CANCELLED\n' if cancelled else 'STATUS:CONFIRMED\n'
	uid = 'UID:'+str(reservation.id)+'\n'
	sequence = 'SEQUENCE:2\n' if cancelled else 'SEQUENCE:0\n'
	priority = 'PRIORITY:5\n' if cancelled else 'PRIORITY:0\n'
	now = datetime.utcnow().strftime('%Y%m%dT%H%M%SZ')
	start = reservation.start.astimezone(pytz.utc).strftime('%Y%m%dT%H%M%SZ')
	end = reservation.end.astimezone(pytz.utc).strftime('%Y%m%dT%H%M%SZ')
	reservation_name = reservation.reservation_item.name
	lines = ['BEGIN:VCALENDAR\n', 'VERSION:2.0\n', method, 'BEGIN:VEVENT\n', uid, sequence, priority, f'DTSTAMP:{now}\n', f'DTSTART:{start}\n', f'DTEND:{end}\n', f'ATTENDEE;CN="{reservation.user.get_name()}";RSVP=TRUE:mailto:{reservation.user.email}\n', f'ORGANIZER;CN="{reservation_organizer}":mailto:{reservation_organizer_email}\n', f'SUMMARY:[{site_title}] {reservation_name} Reservation\n', status, 'END:VEVENT\n', 'END:VCALENDAR\n']
	ics = io.StringIO('')
	ics.writelines(lines)
	ics.seek(0)

	attachment = create_email_attachment(ics, maintype='text', subtype='calendar', method=method_name)
	return attachment


@login_required
@require_GET
@permission_required("NEMO.trigger_timed_services", raise_exception=True)
def create_closure_alerts(request):
	return do_create_closure_alerts()


def do_create_closure_alerts():
	future_times = ClosureTime.objects.filter(closure__alert_days_before__isnull=False, end_time__gt=timezone.now())
	for closure_time in future_times:
		create_alert_for_closure_time(closure_time)
	for closure in Closure.objects.filter(notify_managers_last_occurrence=True):
		closure_time_ending = ClosureTime.objects.filter(closure=closure).latest("end_time")
		if as_timezone(closure_time_ending.end_time).date() == timezone.now().date():
			email_last_closure_occurrence(closure_time_ending)
	return HttpResponse()


def create_alert_for_closure_time(closure_time: ClosureTime):
	# Create alerts a week before their debut time (closure start - days before) at the latest
	now = timezone.now()
	next_week = now + timedelta(weeks=1)
	closure = closure_time.closure
	alert_start = closure_time.start_time - timedelta(days=closure.alert_days_before)
	time_ready = alert_start <= next_week
	if time_ready:
		# Check if there is already an alert with the same title ending at the closure end time
		alert_already_exist = Alert.objects.filter(title=closure.name, expiration_time=closure_time.end_time).exists()
		if not alert_already_exist:
			areas = Area.objects.filter(id__in=distinct_qs_value_list(closure.physical_access_levels.all(), "area"))
			dictionary = {
				"closure_time": closure_time,
				"name": closure.name,
				"staff_absent": closure.staff_absent,
				"start_time": closure_time.start_time,
				"end_time": closure_time.end_time,
				"areas": areas,
			}
			contents = Template(closure.alert_template).render(Context(dictionary)) if closure.alert_template else None
			Alert.objects.create(
				title=closure.name,
				contents=contents,
				category="Closure",
				debut_time=alert_start,
				expiration_time=closure_time.end_time,
			)


def email_last_closure_occurrence(closure_time):
	facility_manager_emails = User.objects.filter(is_active=True, is_facility_manager=True).values_list(
		"email", flat=True
	)
	message = f"""
Dear facility manager,<br>
This email is to inform you that today was the last occurrence for the {closure_time.closure.name} facility closure.
<br><br>Go to NEMO -> Detailed administration -> Closures to add more times if needed.
"""
	send_mail(
		subject=f"Last {closure_time.closure.name} occurrence",
		content=message,
		from_email=settings.SERVER_EMAIL,
		to=facility_manager_emails,
		email_category=EmailCategory.SYSTEM,
	)



from datetime import datetime, timedelta
from logging import getLogger

from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import get_object_or_404, redirect, render
from django.template import Context, Template
from django.urls import reverse
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from django.views.decorators.http import require_GET, require_http_methods

from NEMO.forms import TemporaryPhysicalAccessRequestForm
from NEMO.models import PhysicalAccessLevel, TemporaryPhysicalAccess, TemporaryPhysicalAccessRequest, User
from NEMO.utilities import (
	EmailCategory,
	beginning_of_the_day,
	bootstrap_primary_color,
	format_datetime,
	quiet_int,
	send_mail,
)
from NEMO.views.customization import get_customization, get_media_file_contents, set_customization
from NEMO.views.notifications import create_access_request_notification, delete_notification, get_notifications

access_request_logger = getLogger(__name__)


@login_required
@require_GET
def access_requests(request):
	mark_requests_expired()
	user: User = request.user
	status = TemporaryPhysicalAccessRequest.Status
	max_requests = quiet_int(get_customization("access_requests_display_max"), None)
	physical_access_requests = TemporaryPhysicalAccessRequest.objects.filter(deleted=False)
	physical_access_requests = physical_access_requests.order_by("-end_time")
	if not user.is_facility_manager and not user.is_staff:
		# For some reason doing an "or" filtering with manytomany field returns duplicates, and using distinct() returns nothing...
		other_user_physical_access_requests = physical_access_requests.filter(other_users__in=[user]).distinct()
		physical_access_requests = physical_access_requests.filter(Q(creator=user) | Q(id__in=other_user_physical_access_requests))
	dictionary = {
		"pending_access_requests": physical_access_requests.filter(status=status.PENDING).order_by("start_time"),
		"approved_access_requests": physical_access_requests.filter(status=status.APPROVED)[:max_requests],
		"denied_access_requests": physical_access_requests.filter(status=status.DENIED)[:max_requests],
		"expired_access_requests": physical_access_requests.filter(status=status.EXPIRED)[:max_requests],
		"access_requests_description": get_customization("access_requests_description"),
		"access_request_notifications": get_notifications(
			request.user, TemporaryPhysicalAccessRequest, delete=not user.is_facility_manager
		),
	}
	return render(request, "requests/access_requests/access_requests.html", dictionary)


@login_required
@require_http_methods(["GET", "POST"])
def create_access_request(request, request_id=None):
	user: User = request.user
	try:
		access_request = TemporaryPhysicalAccessRequest.objects.get(id=request_id)
	except TemporaryPhysicalAccessRequest.DoesNotExist:
		access_request = None

	dictionary = {
		"physical_access_levels": PhysicalAccessLevel.objects.filter(allow_user_request=True),
		"other_users": User.objects.filter(is_active=True).exclude(id=user.id),
	}

	if request.method == "POST":
		# some extra validation needs to be done here because it depends on the user
		edit = bool(access_request)
		errors = []
		if edit:
			if access_request.deleted:
				errors.append("You are not allowed to edit expired or deleted requests.")
			if access_request.status != TemporaryPhysicalAccessRequest.Status.PENDING:
				errors.append("Only pending requests can be modified.")
			if access_request.creator != user and not user.is_facility_manager:
				errors.append("You are not allowed to edit a request you didn't create.")

		form = TemporaryPhysicalAccessRequestForm(
			request.POST, instance=access_request, initial={"creator": access_request.creator if edit else user}
		)

		# add errors to the form for better display
		for error in errors:
			form.add_error(None, error)

		cleaned_data = form.clean()
		if cleaned_data and not user.is_facility_manager and cleaned_data.get("start_time") < timezone.now():
			form.add_error("start_time", "The start time must be later than the current time")

		if form.is_valid():
			if not edit:
				form.instance.creator = user
			if edit and user.is_facility_manager:
				decision = [state for state in ["approve_request", "deny_request"] if state in request.POST]
				if decision:
					if next(iter(decision)) == "approve_request":
						access_request.status = TemporaryPhysicalAccessRequest.Status.APPROVED
						create_temporary_access(access_request)
					else:
						access_request.status = TemporaryPhysicalAccessRequest.Status.DENIED
					access_request.reviewer = user

			form.instance.last_updated_by = user
			new_access_request = form.save()
			create_access_request_notification(new_access_request)
			if edit:
				# remove notification for current user and other facility managers
				delete_notification(TemporaryPhysicalAccessRequest, new_access_request.id, [user])
				if user.is_facility_manager:
					managers = User.objects.filter(is_active=True, is_facility_manager=True)
					delete_notification(TemporaryPhysicalAccessRequest, new_access_request.id, managers)
			send_request_received_email(request, new_access_request, edit)
			return redirect("user_requests", "access")
		else:
			dictionary["form"] = form
			return render(request, "requests/access_requests/access_request.html", dictionary)
	else:
		form = TemporaryPhysicalAccessRequestForm(instance=access_request)
		dictionary["form"] = form
		return render(request, "requests/access_requests/access_request.html", dictionary)


@login_required
@require_GET
def delete_access_request(request, request_id):
	access_request = get_object_or_404(TemporaryPhysicalAccessRequest, id=request_id)

	if access_request.creator != request.user:
		return HttpResponseBadRequest("You are not allowed to delete a request you didn't create.")
	if access_request and access_request.status != TemporaryPhysicalAccessRequest.Status.PENDING:
		return HttpResponseBadRequest("You are not allowed to delete a request that is not pending.")

	access_request.deleted = True
	access_request.save(update_fields=["deleted"])
	delete_notification(TemporaryPhysicalAccessRequest, access_request.id)
	return redirect("user_requests", "access")


def create_temporary_access(access_request: TemporaryPhysicalAccessRequest):
	for user in access_request.creator_and_other_users():
		TemporaryPhysicalAccess.objects.create(
			user=user,
			physical_access_level=access_request.physical_access_level,
			start_time=access_request.start_time,
			end_time=access_request.end_time,
		)


def mark_requests_expired():
	for expired_request in TemporaryPhysicalAccessRequest.objects.filter(
			status=TemporaryPhysicalAccessRequest.Status.PENDING, deleted=False, end_time__lt=timezone.now()
	):
		delete_notification(TemporaryPhysicalAccessRequest, expired_request.id)
		expired_request.status = TemporaryPhysicalAccessRequest.Status.EXPIRED
		expired_request.save(update_fields=["status"])


def send_request_received_email(request, access_request: TemporaryPhysicalAccessRequest, edit):
	user_office_email = get_customization("user_office_email_address")
	access_request_notification_email = get_media_file_contents("access_request_notification_email.html")
	if user_office_email and access_request_notification_email:
		facility_manager_emails = User.objects.filter(is_active=True, is_facility_manager=True).values_list(
			"email", flat=True
		)
		ccs = tuple(
			[e for e in [*access_request.other_users.values_list("email", flat=True), *facility_manager_emails] if e]
		)
		ccs += (user_office_email,)
		status = (
			"approved"
			if access_request.status == TemporaryPhysicalAccessRequest.Status.APPROVED
			else "denied"
			if access_request.status == TemporaryPhysicalAccessRequest.Status.DENIED
			else "updated"
			if edit
			else "received"
		)
		absolute_url = request.build_absolute_uri(reverse("user_requests", kwargs={"tab": "access"}))
		color_type = "success" if status == "approved" else "danger" if status == "denied" else "info"
		message = Template(access_request_notification_email).render(
			Context(
				{
					"template_color": bootstrap_primary_color(color_type),
					"access_request": access_request,
					"status": status,
					"access_requests_url": absolute_url,
				}
			)
		)
		send_mail(
			subject=f"Your access request for the {access_request.physical_access_level.area} has been {status}",
			content=message,
			from_email=user_office_email,
			to=[access_request.creator.email],
			cc=ccs,
			email_category=EmailCategory.ACCESS_REQUESTS,
		)


@login_required
@permission_required("NEMO.trigger_timed_services", raise_exception=True)
@require_GET
def email_weekend_access_notification(request):
	return send_email_weekend_access_notification()


def send_email_weekend_access_notification():
	"""
		Sends a weekend access email to the addresses set in customization with the template provided.
		The email is sent when the first request (each week) that includes weekend access is approved.
		If no weekend access requests are made by the given time on the cutoff day (if set), a no access email is sent.
	"""
	try:
		user_office_email = get_customization("user_office_email_address")
		email_to = get_customization("weekend_access_notification_emails")
		access_contents = get_media_file_contents("weekend_access_email.html")
		if user_office_email and email_to and access_contents:
			process_weekend_access_notification(user_office_email, email_to, access_contents)
	except Exception as error:
		access_request_logger.error(error)
	return HttpResponse()


def process_weekend_access_notification(user_office_email, email_to, access_contents):
	today = datetime.today()
	beginning_of_the_week = beginning_of_the_day(today - timedelta(days=today.weekday()))
	cutoff_day = get_customization("weekend_access_notification_cutoff_day")
	cutoff_hour = get_customization("weekend_access_notification_cutoff_hour")
	# Set the cutoff in actual datetime format
	cutoff_datetime = None
	if cutoff_hour.isdigit() and cutoff_day and cutoff_day.isdigit():
		cutoff_datetime = (beginning_of_the_week + timedelta(days=int(cutoff_day))).replace(hour=int(cutoff_hour))

	end_of_the_week = beginning_of_the_week + timedelta(weeks=1)
	beginning_of_the_weekend = beginning_of_the_week + timedelta(days=5)

	# Approved access request that include weekend time do overlap with weekend date interval.
	approved_status = TemporaryPhysicalAccessRequest.Status.APPROVED
	approved_weekend_access_requests = TemporaryPhysicalAccessRequest.objects.filter(
		deleted=False, status=approved_status
	)
	approved_weekend_access_requests = approved_weekend_access_requests.exclude(start_time__gte=end_of_the_week)
	approved_weekend_access_requests = approved_weekend_access_requests.exclude(end_time__lte=beginning_of_the_weekend)

	cutoff_time_passed = cutoff_datetime and timezone.now() >= cutoff_datetime
	last_sent = get_customization("weekend_access_notification_last_sent")
	last_sent_datetime = parse_datetime(last_sent) if last_sent else None
	if (
			(not last_sent_datetime or last_sent_datetime < beginning_of_the_week)
			and access_contents
			and approved_weekend_access_requests.exists()
			and not cutoff_time_passed
	):
		send_weekend_email_access(True, user_office_email, email_to, access_contents, beginning_of_the_week)
		set_customization("weekend_access_notification_last_sent", str(timezone.now()))
	if access_contents and cutoff_datetime and not approved_weekend_access_requests.exists():
		is_cutoff = today.weekday() == int(cutoff_day) and cutoff_datetime.hour == timezone.localtime().hour
		if is_cutoff:
			send_weekend_email_access(False, user_office_email, email_to, access_contents, beginning_of_the_week)


def send_weekend_email_access(access, user_office_email, email_to, contents, beginning_of_the_week):
	facility_name = get_customization("facility_name")
	manager_emails = User.objects.filter(is_active=True, is_facility_manager=True).values_list("email", flat=True)
	recipients = tuple([e for e in email_to.split(",") if e])
	ccs = tuple([e for e in manager_emails if e])
	ccs += (user_office_email,)

	sat = format_datetime(beginning_of_the_week + timedelta(days=5), "SHORT_DATE_FORMAT", as_current_timezone=False)
	sun = format_datetime(beginning_of_the_week + timedelta(days=6), "SHORT_DATE_FORMAT", as_current_timezone=False)

	subject = f"{'NO w' if not access else 'W'}eekend access for the {facility_name} {sat} - {sun}"
	message = Template(contents).render(Context({"weekend_access": access}))
	send_mail(
		subject=subject,
		content=message,
		from_email=user_office_email,
		to=recipients,
		cc=ccs,
		email_category=EmailCategory.ACCESS_REQUESTS,
	)
